import logging
import time
from collections import defaultdict
from typing import Sequence, cast, Literal

from hhd.controller import (
    Consumer,
    Event,
    Producer,
    TouchpadCorrectionType,
    correct_touchpad,
)
from hhd.controller.lib.common import encode_axis, set_button
from hhd.controller.lib.uhid import BUS_BLUETOOTH, BUS_USB, UhidDevice
from hhd.controller.lib.ccache import ControllerCache

from .const import (
    DS5_BT_AXIS_MAP,
    DS5_BT_BTN_MAP,
    DS5_EDGE_COUNTRY,
    DS5_EDGE_DELTA_TIME_NS,
    DS5_EDGE_DESCRIPTOR_BT,
    DS5_EDGE_DESCRIPTOR_USB,
    DS5_EDGE_MAX_REPORT_FREQ,
    DS5_EDGE_MIN_REPORT_FREQ,
    DS5_EDGE_NAME,
    DS5_EDGE_PRODUCT,
    DS5_EDGE_REPORT_PAIRING,
    DS5_EDGE_REPORT_PAIRING_ID,
    DS5_EDGE_STOCK_REPORTS,
    DS5_EDGE_TOUCH_HEIGHT,
    DS5_EDGE_TOUCH_WIDTH,
    DS5_EDGE_VERSION,
    DS5_FEATURE_CRC32_SEED,
    DS5_INPUT_CRC32_SEED,
    DS5_INPUT_REPORT_BT_OFS,
    DS5_INPUT_REPORT_USB_OFS,
    DS5_NAME,
    DS5_PRODUCT,
    DS5_USB_AXIS_MAP,
    DS5_USB_BTN_MAP,
    DS5_NAME_LEFT,
    DS5_VENDOR,
    patch_dpad_val,
    prefill_ds5_report,
    sign_crc32_append,
    sign_crc32_inplace,
)

REPORT_MAX_DELAY = 1 / DS5_EDGE_MIN_REPORT_FREQ
REPORT_MIN_DELAY = 1 / DS5_EDGE_MAX_REPORT_FREQ
DS5_EDGE_MIN_TIMESTAMP_INTERVAL = 1500
MAX_IMU_SYNC_DELAY = 2

logger = logging.getLogger(__name__)

_cache = ControllerCache()
_cache_left = ControllerCache()


class Dualsense(Producer, Consumer):
    @staticmethod
    def close_cached():
        _cache.close()
        _cache_left.close()

    def __init__(
        self,
        touchpad_method: TouchpadCorrectionType = "crop_end",
        edge_mode: bool = True,
        use_bluetooth: bool = True,
        fake_timestamps: bool = False,
        enable_touchpad: bool = True,
        enable_rgb: bool = True,
        sync_gyro: bool = False,
        flip_z: bool = True,
        paddles_to_clicks: Literal["disabled", "top", "bottom"] = "disabled",
        controller_id: int = 0,
        left_motion: bool = False,
        cache: bool = False,
    ) -> None:
        self.available = False
        self.report = None
        self.dev = None
        self.start = 0
        self.use_bluetooth = use_bluetooth
        self.edge_mode = edge_mode
        self.fake_timestamps = fake_timestamps
        self.touchpad_method: TouchpadCorrectionType = touchpad_method
        self.enable_touchpad = enable_touchpad
        self.enable_rgb = enable_rgb
        self.sync_gyro = sync_gyro
        self.flip_z = flip_z
        self.paddles_to_clicks = paddles_to_clicks
        self.controller_id = controller_id
        self.left_motion = left_motion
        self.cache = cache
        self.last_imu_ts = 0

        self.ofs = (
            DS5_INPUT_REPORT_BT_OFS if use_bluetooth else DS5_INPUT_REPORT_USB_OFS
        )
        self.axis_map = DS5_BT_AXIS_MAP if use_bluetooth else DS5_USB_AXIS_MAP
        self.btn_map = DS5_BT_BTN_MAP if use_bluetooth else DS5_USB_BTN_MAP

    def open(self) -> Sequence[int]:
        self.available = False
        self.report = bytearray(prefill_ds5_report(self.use_bluetooth))

        cached = cast(
            Dualsense | None, _cache_left.get() if self.left_motion else _cache.get()
        )

        # Use cached controller to avoid disconnects
        self.dev = None
        if cached:
            if (
                self.edge_mode == cached.edge_mode
                and self.use_bluetooth == cached.use_bluetooth
                and self.controller_id == cached.controller_id
            ):
                logger.warning(
                    f"Using cached controller node for Dualsense {'left motions device' if self.left_motion else 'controller'}."
                )
                self.dev = cached.dev
                if self.dev and self.dev.fd:
                    self.fd = self.dev.fd
            else:
                logger.warning(
                    f"Throwing away cached Dualsense for {'left motions device' if self.left_motion else 'controller'}."
                )
                cached.close(True, in_cache=True)
        name = (
            (DS5_EDGE_NAME if self.edge_mode else DS5_NAME)
            if not self.left_motion
            else DS5_NAME_LEFT
        )
        if not self.dev:
            self.dev = UhidDevice(
                vid=DS5_VENDOR,
                pid=DS5_EDGE_PRODUCT if self.edge_mode else DS5_PRODUCT,
                bus=BUS_BLUETOOTH if self.use_bluetooth else BUS_USB,
                version=DS5_EDGE_VERSION,
                country=DS5_EDGE_COUNTRY,
                name=name,
                report_descriptor=(
                    DS5_EDGE_DESCRIPTOR_BT
                    if self.use_bluetooth
                    else DS5_EDGE_DESCRIPTOR_USB
                ),
            )
            self.fd = self.dev.open()

        self.touch_correction = correct_touchpad(
            DS5_EDGE_TOUCH_WIDTH, DS5_EDGE_TOUCH_HEIGHT, 1, self.touchpad_method
        )

        self.state: dict = defaultdict(lambda: 0)
        self.rumble = False
        self.touchpad_touch = False
        curr = time.perf_counter()
        self.touchpad_down = curr
        self.last_imu = curr
        self.imu_failed = False
        self.start = time.perf_counter()

        logger.info(f"Starting '{name.decode()}'.")
        assert self.fd
        return [self.fd]

    def close(self, exit: bool, in_cache: bool = False) -> bool:
        if not in_cache and self.cache and time.perf_counter() - self.start:
            logger.warning(
                f"Caching Dualsense {'left motions device' if self.left_motion else 'controller'} to avoid reconnection."
            )
            if self.left_motion:
                _cache_left.add(self)
            else:
                _cache.add(self)
        elif self.dev:
            self.dev.send_destroy()
            self.dev.close()
            self.dev = None
            self.fd = None

        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if self.fd not in fds:
            return []

        # Process queued events
        out: Sequence[Event] = []
        assert self.dev
        while ev := self.dev.read_event():
            match ev["type"]:
                case "open":
                    self.available = True
                case "close":
                    self.available = False
                case "get_report":
                    if ev["rnum"] in DS5_EDGE_STOCK_REPORTS:
                        num = ev["rnum"]
                        if num == DS5_EDGE_REPORT_PAIRING_ID:
                            # Customize pairing report to have per-controller
                            # calibration
                            rep = DS5_EDGE_REPORT_PAIRING(self.controller_id)
                        else:
                            rep = DS5_EDGE_STOCK_REPORTS[num]

                        if self.use_bluetooth:
                            rep = sign_crc32_append(rep, DS5_FEATURE_CRC32_SEED)
                        self.dev.send_get_report_reply(ev["id"], 0, rep)
                    else:
                        logger.warning(
                            f"DS5: Received get_report with the id (uknown): {ev['rnum']}"
                        )
                case "set_report":
                    logger.warning(
                        f"DS5: Received set_report with the id (uknown): {ev['rnum']}"
                    )
                case "output":
                    invalid = False
                    # Check report num
                    if ev["report"] != 0x01:
                        invalid = True
                    # Check report ids depending on modes
                    if not self.use_bluetooth and ev["data"][0] != 0x02:
                        invalid = True
                    if self.use_bluetooth and ev["data"][0] != 0x31:
                        invalid = True

                    if invalid:
                        logger.warning(
                            f"DS5: Received uknown output report with the following data:\n{ev['report']}: {ev['data'].hex()}"
                        )
                        continue

                    rep = ev["data"]
                    if self.use_bluetooth:
                        # skip seq_tag, tag sent by bluetooth report
                        # rest is the same

                        # If the first byte is the sequence byte, it will be
                        # from 0x00 to 0xF0. Otherwise, for sdl that does not
                        # have it it will be 0x02.
                        # Only the kernel appends the sequence byte
                        # SDL does not
                        if rep[1] == 0x02:
                            rep = rep[0:1] + rep[2:]
                        else:
                            rep = rep[0:1] + rep[3:]

                    flag0 = rep[1]
                    flag1 = rep[2]
                    flag2 = rep[39]
                    if self.enable_rgb and (
                        flag1 & 4
                    ):  # DS_OUTPUT_VALID_FLAG1_LIGHTBAR_CONTROL_ENABLE
                        # Led data is being set
                        led_brightness = rep[43]
                        player_leds = rep[44]
                        red = rep[45]
                        green = rep[46]
                        blue = rep[47]
                        if red == 0 and green == 0 and blue == 128:
                            # Skip playstation driver initialization
                            continue
                        if red == 0 and green == 0 and blue == 64:
                            # Skip SDL led initialization
                            continue
                        if red == 64 and green == 0 and blue == 0:
                            # Skip rare SDL led initialization that is offset
                            continue
                        logger.info(f"Changing leds to RGB: {red} {green} {blue}")

                        # Crunch lower values since steam is bugged
                        if red < 3 and green < 3 and blue < 3:
                            red = 0
                            green = 0
                            blue = 0

                        out.append(
                            {
                                "type": "led",
                                "code": "main",
                                "mode": "solid",
                                # "brightness": led_brightness / 63
                                # if led_brightness
                                # else 1,
                                "initialize": False,
                                "direction": "left",
                                "speed": 0,
                                "brightness": 1,
                                "speedd": "high",
                                "brightnessd": "high",
                                "red": red,
                                "blue": blue,
                                "green": green,
                                "red2": 0, # disable for OXP
                                "blue2": 0,
                                "green2": 0,
                                "oxp": None,
                            }
                        )
                    # elif (rep[39] & 2) and (rep[42] & 2):
                    #     # flag2 is DS_OUTPUT_VALID_FLAG2_LIGHTBAR_SETUP_CONTROL_ENABLE
                    #     # lightbar_setup is DS_OUTPUT_LIGHTBAR_SETUP_LIGHT_OUT
                    #     # FIXME: Disable for now to avoid hid_playstation messing
                    #     # with the leds
                    #     out.append(
                    #         {
                    #             "type": "led",
                    #             "code": "main",
                    #             "mode": "disable",
                    #             "brightness": 0,
                    #             "speed": 0,
                    #             "red": 0,
                    #             "blue": 0,
                    #             "green": 0,
                    #         }
                    #     )
                    #     pass

                    # Rumble
                    # Flag 1
                    # Death stranding uses 0x40 to turn on vibration
                    # SDL uses 0x02 to disable audio haptics
                    # old version used flag0 & 0x02
                    # Initial compatibility rumble is flag0 0x01
                    # Improved is flag2 0x04
                    if flag0 & 0x01 or flag2 & 0x04:
                        right = rep[3]
                        left = rep[4]

                        # If vibration mode is in flag0 use different scale
                        scale = 2 if flag0 & 0x01 else 1

                        out.append(
                            {
                                "type": "rumble",
                                "code": "main",
                                # For some reason goes to 127
                                "strong_magnitude": left / 255 * scale,
                                "weak_magnitude": right / 255 * scale,
                            }
                        )
                        self.rumble = True
                    elif self.rumble:
                        self.rumble = False
                        out.append(
                            {
                                "type": "rumble",
                                "code": "main",
                                "strong_magnitude": 0,
                                "weak_magnitude": 0,
                            }
                        )
                case _:
                    logger.debug(f"Received unhandled report:\n{ev}")
        return out

    def consume(self, events: Sequence[Event]):
        assert self.dev and self.report
        # To fix gyro to mouse in latest steam
        # only send updates when gyro sends a timestamp
        send = not self.sync_gyro
        curr = time.perf_counter()

        new_rep = bytearray(self.report)
        for ev in events:
            code = ev["code"]
            match ev["type"]:
                case "axis":
                    if not self.enable_touchpad and code.startswith("touchpad"):
                        continue
                    if self.left_motion:
                        # Only left keep imu events for left motion
                        if (
                            "left_gyro_" in code
                            or "left_accel_" in code
                            or "left_imu_" in code
                        ):
                            code = code.replace("left_", "")
                        else:
                            continue
                    if code in self.axis_map:
                        if self.flip_z and code == "gyro_z":
                            ev["value"] = -ev["value"]
                        try:
                            encode_axis(new_rep, self.axis_map[code], ev["value"])
                        except Exception:
                            logger.warning(
                                f"Encoding '{ev['code']}' with {ev['value']} overflowed."
                            )
                    # DPAD is weird
                    match code:
                        case "hat_x":
                            self.state["hat_x"] = ev["value"]
                            patch_dpad_val(
                                new_rep,
                                self.ofs,
                                self.state["hat_x"],
                                self.state["hat_y"],
                            )
                        case "hat_y":
                            self.state["hat_y"] = ev["value"]
                            patch_dpad_val(
                                new_rep,
                                self.ofs,
                                self.state["hat_x"],
                                self.state["hat_y"],
                            )
                        case "touchpad_x":
                            tc = self.touch_correction
                            x = int(
                                min(max(ev["value"], tc.x_clamp[0]), tc.x_clamp[1])
                                * tc.x_mult
                                + tc.x_ofs
                            )
                            new_rep[self.ofs + 33] = x & 0xFF
                            new_rep[self.ofs + 34] = (new_rep[self.ofs + 34] & 0xF0) | (
                                x >> 8
                            )
                        case "touchpad_y":
                            tc = self.touch_correction
                            y = int(
                                min(max(ev["value"], tc.y_clamp[0]), tc.y_clamp[1])
                                * tc.y_mult
                                + tc.y_ofs
                            )
                            new_rep[self.ofs + 34] = (new_rep[self.ofs + 34] & 0x0F) | (
                                (y & 0x0F) << 4
                            )
                            new_rep[self.ofs + 35] = y >> 4
                        case "gyro_ts" | "accel_ts" | "imu_ts":
                            send = True
                            self.last_imu = time.perf_counter()
                            self.last_imu_ts = ev["value"]
                            new_rep[self.ofs + 27 : self.ofs + 31] = int(
                                ev["value"] / DS5_EDGE_DELTA_TIME_NS
                            ).to_bytes(8, byteorder="little", signed=False)[:4]
                case "button":
                    if self.left_motion:
                        # skip buttons for left motion
                        continue
                    if not self.enable_touchpad and code.startswith("touchpad"):
                        continue
                    if (self.paddles_to_clicks == "top" and code == "extra_l1") or (
                        self.paddles_to_clicks == "bottom" and code == "extra_l2"
                    ):
                        # Place finger on correct place and click
                        new_rep[self.ofs + 33] = 0x80
                        new_rep[self.ofs + 34] = 0x01
                        new_rep[self.ofs + 35] = 0x20
                        # Replace code with click
                        ev = {**ev, "code": "touchpad_left"}
                        code = "touchpad_left"
                    if (self.paddles_to_clicks == "top" and code == "extra_r1") or (
                        self.paddles_to_clicks == "bottom" and code == "extra_r2"
                    ):
                        # Place finger on correct place and click
                        new_rep[self.ofs + 33] = 0x00
                        new_rep[self.ofs + 34] = 0x06
                        new_rep[self.ofs + 35] = 0x20
                        # Replace code with click
                        ev = {**ev, "code": "touchpad_left"}
                        code = "touchpad_left"

                    if code in self.btn_map:
                        set_button(new_rep, self.btn_map[code], ev["value"])

                    # Fix touchpad click requiring touch
                    if code == "touchpad_touch":
                        self.touchpad_touch = ev["value"]
                    if code == "touchpad_left":
                        set_button(
                            new_rep,
                            self.btn_map["touchpad_touch"],
                            ev["value"] or self.touchpad_touch,
                        )
                    # Also add right click
                    if code == "touchpad_right":
                        set_button(
                            new_rep,
                            self.btn_map["touchpad_touch"],
                            ev["value"] or self.touchpad_touch,
                        )
                        set_button(
                            new_rep,
                            self.btn_map["touchpad_touch2"],
                            ev["value"],
                        )

                case "configuration":
                    if self.left_motion:
                        continue
                    match code:
                        case "touchpad_aspect_ratio":
                            self.aspect_ratio = cast(float, ev["value"])
                            self.touch_correction = correct_touchpad(
                                DS5_EDGE_TOUCH_WIDTH,
                                DS5_EDGE_TOUCH_HEIGHT,
                                self.aspect_ratio,
                                self.touchpad_method,
                            )
                        case "is_attached":
                            new_rep[self.ofs + 52] = (new_rep[self.ofs + 52] & 0x0F) | (
                                0x10 if ev["value"] else 0x00
                            )
                        case "battery":
                            new_rep[self.ofs + 52] = (new_rep[self.ofs + 52] & 0xF0) | (
                                max(ev["value"] // 10, 0)
                            )

        # Cache
        # Caching can cause issues since receivers expect reports
        # at least a couple of times per second
        # if new_rep == self.report and not self.fake_timestamps:
        #     return
        self.report = new_rep

        # If the IMU breaks, smoothly re-enable the controller
        failover = self.last_imu + MAX_IMU_SYNC_DELAY < curr
        if self.sync_gyro and failover and not self.imu_failed:
            self.imu_failed = True
            logger.error(
                f"IMU Did not send information for {MAX_IMU_SYNC_DELAY}s. Disabling Gyro Sync."
            )

        if self.fake_timestamps or failover:
            new_rep[self.ofs + 27 : self.ofs + 31] = int(
                time.perf_counter_ns() / DS5_EDGE_DELTA_TIME_NS
            ).to_bytes(8, byteorder="little", signed=False)[:4]

        #
        # Send report
        #
        # Sequence number
        if new_rep[self.ofs + 6] < 255:
            new_rep[self.ofs + 6] += 1
        else:
            new_rep[self.ofs + 6] = 0

        if self.use_bluetooth:
            sign_crc32_inplace(self.report, DS5_INPUT_CRC32_SEED)
        if send or failover:
            self.dev.send_input_report(self.report)



#!/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024      Antheas Kapenekakis
# Copyright (c) 2012-2017 Benjamin Tissoires <benjamin.tissoires@gmail.com>
# Copyright (c) 2012-2017 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import copy
import enum
import functools
import itertools
import logging
import re
import os
import sys
from collections.abc import ItemsView, Iterable
from typing import (
    IO,
    Annotated,
    Any,
    Dict,
    Final,
    Hashable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeAlias,
    Union,
    cast,
)

_Type = Type


class ValueRange(NamedTuple):
    min: int
    max: int


DATA_DIRNAME = "hid_data"
SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
DATA_DIR = os.path.join(SCRIPT_DIR, DATA_DIRNAME)

U8 = Annotated[int, ValueRange(0, 0xFF)]
U16 = Annotated[int, ValueRange(0, 0xFFFF)]
U32 = Annotated[int, ValueRange(0, 0xFFFFFFFF)]

dict_items_usage: TypeAlias = ItemsView[U16, "HidUsage"]
dict_items_usagePage: TypeAlias = ItemsView[U16, "HidUsagePage"]


@functools.total_ordering
class HidUsage(Hashable):
    """
    A HID Usage entry as defined in the HID Usage Tablets. ::

        > usage_page = hidtools.hut.HUT[0x01]  # Generic Desktop
        > usage = usage_page[0x02]
        > print(usage.usage)
        2
        > print(usage)
        Mouse
        > print(usage.name)
        Mouse

    :param HidUsagePage usage_page: the Usage Page this Usage belongs to
    :param int usage: the 16-bit Usage assigned by the HID Usage Tables
    :param str name: the usage_name

    .. attribute:: usage

        the 16-bit Usage assigned by the HId Usage Tables

    .. attribute:: name

        the semantic name for this Usage

    .. attribute:: usage_page

        the :class:`HidUsagePage` this Usage belongs to

    """

    def __init__(
        self: "HidUsage", usage_page: "HidUsagePage", usage: U16, name: str
    ) -> None:
        self.usage_page = usage_page
        self.usage = usage
        self.name = name

    # Route everything down to the name, this way we basically behave like a
    # string
    def __getattr__(self: "HidUsage", attr: str) -> Any:
        return getattr(self.name, attr)

    def __repr__(self: "HidUsage") -> str:
        return self.name

    def __hash__(self: "HidUsage") -> int:
        return hash(self.name)

    def __str__(self: "HidUsage") -> str:
        return self.name

    def __eq__(self: "HidUsage", other: object) -> bool:
        if isinstance(other, HidUsage):
            return self.name == other.name
        elif not isinstance(other, str):
            return NotImplemented
        return self.name == other

    def __lt__(self: "HidUsage", other: object) -> bool:
        if isinstance(other, HidUsage):
            return self.name < other.name
        elif not isinstance(other, str):
            return NotImplemented
        return self.name < other


class HidUsagePage(object):
    """
    A dictionary of HID Usages in the form ``{usage: usage_name}``,
    representing all Usages in this Usage Page.

    A HID Usage is named semantical identifier that describe how a given
    field in a HID report is to be used. A Usage Page is a logical grouping
    of those identifiers, e.g. "Generic Desktop", "Telephony Devices", or
    "Digitizers".  ::

        > print(usage_page.page_name)
        Generic Desktop
        > print(usage_page.page_id)
        1
        > print(usage_page[0x02])
        Mouse
        > print(usage_page['Mouse'])
        Mouse
        > usage = usage_page.from_name["Mouse"]
        > print(usage.usage)
        2
        > print(usage.name)
        Mouse
        > print(usage)
        Mouse

    .. attribute:: page_id

        The Page ID for this Usage Page, e.g. ``01`` (Generic Desktop)

    .. attribute:: page_name

        The assigned name for this usage Page, e.g. "Generic Desktop"
    """

    def __init__(self: "HidUsagePage") -> None:
        self._usages: Dict[U16, HidUsage] = {}

    def __setitem__(self: "HidUsagePage", key: U16, value: HidUsage) -> None:
        self._usages[key] = value

    def __getitem__(self: "HidUsagePage", key: Union[str, U16, U32]) -> HidUsage:
        if isinstance(key, str):
            return self.from_name[key]

        # extract the usage if we have a 32-bit usage and the page ID
        # matches
        if key > 0xFFFF and key & 0xFFFF0000 == self.page_id << 16:
            key &= 0xFFFF
        return self._usages[key]

    def __delitem__(self: "HidUsagePage", key: U16) -> None:
        del self._usages[key]

    def __iter__(self: "HidUsagePage") -> Iterator[U16]:
        return iter(self._usages)

    def __len__(self: "HidUsagePage") -> int:
        return len(self._usages)

    def __str__(self: "HidUsagePage") -> str:
        return self.page_name

    def __repr__(self: "HidUsagePage") -> str:
        return self.page_name

    def items(self: "HidUsagePage") -> dict_items_usage:
        """
        Iterate over all elements, see :meth:`dict.items`
        """
        return self._usages.items()

    @property
    def page_id(self: "HidUsagePage") -> U16:
        """
        The numerical page ID for this usage page
        """
        return self._page_id

    @page_id.setter
    def page_id(self: "HidUsagePage", page_id: U16) -> None:
        self._page_id = page_id

    @property
    def page_name(self: "HidUsagePage") -> str:
        """
        The assigned name for this Usage Page
        """
        return self._name

    @page_name.setter
    def page_name(self: "HidUsagePage", name: str) -> None:
        self._name = name

    @property
    def from_name(self: "HidUsagePage") -> Dict[str, HidUsage]:
        """
        A dictionary using ``{ name: usage }`` mapping, to look up the
        :class:`HidUsage` based on a name.
        """
        try:
            return self._inverted
        except AttributeError:
            self._inverted: Dict[str, HidUsage] = {}
            for _, v in self.items():
                self._inverted[v.name] = v
            return self._inverted

    @property
    def from_usage(self: "HidUsagePage") -> Dict[U16, HidUsage]:
        """
        A dictionary using ``{ usage: name }`` mapping, to look up the name
        based on a page ID . This is the same as using the object itself.
        """
        return cast(Dict[U16, HidUsage], self)


class HidUsageTable(object):
    """
    This effectively a dictionary of all HID Usages known to man. Or to this
    module at least. This object is a singleton, it is available as
    ``hidtools.hut.HUT``.

    Elements of this dictionary are :class:`HidUsagePage` objects.

    This object is a dictionary, use like this: ::

        > hut = hidtools.hut.HUT
        > print(hut[0x01].page_name)
        Generic Desktop
        > print(hut['Generic Desktop'].page_name)
        Generic Desktop
        > print(hut.usage_pages[0x01].page_name)
        Generic Desktop
        > print(hut.usage_page_names['Generic Desktop'].page_name)
        Generic Desktop
        > print(hut[0x01].page_id)
        1
        > print(hut.usage_page_from_name('Generic Desktop').page_id)
        1
        > print(hut.usage_page_from_page_id(0x01).page_name)
        Generic Desktop
    """

    def __init__(self: "HidUsageTable") -> None:
        self._pages: Dict[U16, HidUsagePage] = {}

    def __setitem__(self: "HidUsageTable", key: U16, value: HidUsagePage) -> None:
        self._pages[key] = value

    def __getitem__(self: "HidUsageTable", key: Union[str, U16]) -> HidUsagePage:
        if isinstance(key, str):
            return self.usage_page_names[key]

        # shift the usage page bits down if we have a 32-bit usage
        if key & 0xFFFF0000 == key:
            key >>= 16
        return self._pages[key]

    def __delitem__(self: "HidUsageTable", key) -> None:
        del self._pages[key]

    def __iter__(self: "HidUsageTable") -> Iterator[HidUsagePage]:
        return iter(self._pages)

    def __len__(self: "HidUsageTable") -> int:
        return len(self._pages)

    def items(self: "HidUsageTable") -> dict_items_usagePage:
        """
        Iterate over all elements, see :meth:`dict.items`
        """
        return self._pages.items()

    @property
    def usage_pages(self: "HidUsageTable") -> Dict[U16, HidUsagePage]:
        """
        A dictionary mapping ``{page_id : object}``. These two are
        equivalent calls: ::

            HUT[0x1]
            HUT.usage_pages[0x1]

        """
        return self._pages

    @property
    def usage_page_names(self: "HidUsageTable") -> Dict[str, HidUsagePage]:
        """
        A dictionary mapping ``{page_name : object}``. These two are
        equivalent calls: ::

            HUT['Generic Desktop']
            HUT.usage_page_names['Generic Desktop']

        """
        return {v.page_name: v for _, v in self.items()}

    def usage_page_from_name(
        self: "HidUsageTable", page_name: str
    ) -> Optional[HidUsagePage]:
        """
        Look up the usage page based on the page name (e.g. "Generic
        Desktop"). This is identical to ::

            self.usage_page_names[page_name]

        except that this function returns ``None`` if the page name is
        unknown.

        :return: the :meth:`HidUsagePage` or None
        """
        try:
            return self[page_name]
        except KeyError:
            return None

    def usage_page_from_page_id(
        self: "HidUsageTable", page_id: U16
    ) -> Optional[HidUsagePage]:
        """
        Look up the usage page based on the page ID. This is identical to ::

                self.usage_pages[page_id]

        except that this function returns ``None`` if the page ID is unknown.

        :return: the :meth:`HidUsagePage` or None
        """
        try:
            return self[page_id]
        except KeyError:
            return None

    @classmethod
    def _parse_usages(cls: Type["HidUsageTable"], f: Iterable[str]) -> HidUsagePage:
        """
        Parse a single HUT file. The file format is a set of lines in three
        formats: ::

            (01)<tab>Usage Page name
            A0<tab>Name
            F0-FF<tab>Reserved for somerange

        All numbers in hex.

        Only one Usage Page per file

        Usages are parsed into a dictionary[number] = name.

        The return value is a single HidUsagePage where page[idx] = idx-name.
        """
        usage_page = None
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            # Usage Page, e.g. '(01)	Generic Desktop'
            if line.startswith("("):
                assert usage_page is None

                r = re.match(r"\((?P<idx>[0-9a-fA-F]+)\)\t(?P<page_name>.+)", line)
                assert r is not None
                usage_page = HidUsagePage()
                usage_page.page_id = r["idx"]  # type: ignore
                usage_page.page_name = r["page_name"]
                continue

            assert usage_page is not None

            # Reserved ranges, e.g  '0B-1F	Reserved'
            # "{:x}-{:x}\t{name}"
            r = re.match(
                r"(?P<start>[0-9a-fA-FxX]+)-(?P<end>[0-9a-fA-FxX]+)\S+(?P<name>.+)",
                line,
            )
            if r:
                if "reserved" not in r["name"].lower():
                    print(line)
                continue

            # Single usage, e.g. 36	Slider
            r = re.match(r"(?P<usage>[0-9a-fA-FxX]+)\S+(?P<name>.+)", line)
            assert r is not None, f'"{line}"'
            if "reserved" in r["name"].lower():
                continue

            u = int(r["usage"], 16)
            usage = HidUsage(usage_page, u, r["name"])

            usage_page[u] = usage

        if usage_page is None:
            raise Exception

        return usage_page

    @classmethod
    def _from_hut_data(cls: Type["HidUsageTable"]) -> "HidUsageTable":
        """
        Return the HID Usage Tables, the keys are the numeric Usage Page and
        the values are the respective :class:`hidtools.HidUsagePage` object.

        ::

            > usages = hidtools.hut.HUT()
            > print(usages[0x01].page_name)
            Generic Desktop
            > print(usages.usage_pages[0x01].page_name)
            Generic Desktop
            > print(usages[0x01].page_id)
            1

        :return: a :class:`hidtools.HidUsageTable` object
        """
        hut = HidUsageTable()
        for filename in os.listdir(DATA_DIR):
            if filename.endswith(".hut"):
                with open(os.path.join(DATA_DIR, filename), "r", encoding="utf-8") as f:
                    try:
                        usage_page = cls._parse_usages(f)
                        hut[usage_page.page_id] = usage_page
                    except:
                        print(filename)
                        raise

        return hut


HUT = HidUsageTable._from_hut_data()
"""
The HID Usage Tables as a :class:`hidtools.HidUsageTable` object,
a dictionary where the keys are the numeric Usage Page and the values are
the respective :class:`hidtools.HidUsagePage` object. ::

    > usages = hidtools.hut.HUT()
    > print(usages[0x01].page_name)
    Generic Desktop
    > print(usages.usage_pages[0x01].page_name)
    Generic Desktop
    > print(usages[0x01].page_id)
    1
"""


class BusType(enum.IntEnum):
    """
    The numerical bus type (``0x3`` for USB, ``0x5`` for Bluetooth, see
        ``linux/input.h``)
    """

    PCI = 0x01
    ISAPNP = 0x02
    USB = 0x03
    HIL = 0x04
    BLUETOOTH = 0x05
    VIRTUAL = 0x06
    ISA = 0x10
    I8042 = 0x11
    XTKBD = 0x12
    RS232 = 0x13
    GAMEPORT = 0x14
    PARPORT = 0x15
    AMIGA = 0x16
    ADB = 0x17
    I2C = 0x18
    HOST = 0x19
    GSC = 0x1A
    ATARI = 0x1B
    SPI = 0x1C
    RMI = 0x1D
    CEC = 0x1E
    INTEL_ISHTP = 0x1F
    AMD_SFH = 0x20


def twos_comp(val, bits):
    """compute the 2's complement of val.

    :param int val:
        the value to compute the two's complement for

    :param int bits:
        size of val in bits
    """
    if bits and (val & (1 << (bits - 1))) != 0:
        val = val - (1 << bits)
    return val


def to_twos_comp(val, bits):
    return val & ((1 << bits) - 1)


logger = logging.getLogger("hidtools.hid")

# mypy is confused by the various .bytes properties, so redefine the bytes type
Bytes: TypeAlias = bytes

hid_items: Final = {
    "Main": {
        "Input": 0b10000000,  # noqa: E203
        "Output": 0b10010000,  # noqa: E203
        "Feature": 0b10110000,  # noqa: E203
        "Collection": 0b10100000,  # noqa: E203
        "End Collection": 0b11000000,  # noqa: E203
    },
    "Global": {
        "Usage Page": 0b00000100,  # noqa: E203
        "Logical Minimum": 0b00010100,  # noqa: E203
        "Logical Maximum": 0b00100100,  # noqa: E203
        "Physical Minimum": 0b00110100,  # noqa: E203
        "Physical Maximum": 0b01000100,  # noqa: E203
        "Unit Exponent": 0b01010100,  # noqa: E203
        "Unit": 0b01100100,  # noqa: E203
        "Report Size": 0b01110100,  # noqa: E203
        "Report ID": 0b10000100,  # noqa: E203
        "Report Count": 0b10010100,  # noqa: E203
        "Push": 0b10100100,  # noqa: E203
        "Pop": 0b10110100,  # noqa: E203
    },  # noqa: E203
    "Local": {
        "Usage": 0b00001000,  # noqa: E203
        "Usage Minimum": 0b00011000,  # noqa: E203
        "Usage Maximum": 0b00101000,  # noqa: E203
        "Designator Index": 0b00111000,  # noqa: E203
        "Designator Minimum": 0b01001000,  # noqa: E203
        "Designator Maximum": 0b01011000,  # noqa: E203
        "String Index": 0b01111000,  # noqa: E203
        "String Minimum": 0b10001000,  # noqa: E203
        "String Maximum": 0b10011000,  # noqa: E203
        "Delimiter": 0b10101000,  # noqa: E203
    },
}

superscripts: Final = {
    "0": "⁰",
    "1": "¹",
    "2": "²",
    "3": "³",
    "4": "⁴",
    "5": "⁵",
    "6": "⁶",
    "7": "⁷",
    "8": "⁸",
    "9": "⁹",
    "-": "⁻",
}


class HidUnit(object):
    """
    A parsed field of a HID Report Descriptor Unit specification.

    .. attribute:: units

        A dict of { unit: exponent } of the applicable units.
        Where the Unit is ``None``, the return value is ``None``.

    .. attribute:: system

        The system the units belong to, one of :class:`HidUnit.System`.

    """

    NONE: Final = cast("HidUnit", None)  # For Unit(None), makes the code more obvious

    class System(enum.IntEnum):
        NONE = 0
        SI_LINEAR = 1
        SI_ROTATION = 2
        ENGLISH_LINEAR = 3
        ENGLISH_ROTATION = 4

        @classmethod
        def _stringmap(cls: _Type["HidUnit.System"]) -> Dict["HidUnit.System", str]:
            return {
                HidUnit.System.NONE: "None",
                HidUnit.System.SI_LINEAR: "SILinear",
                HidUnit.System.SI_ROTATION: "SIRotation",
                HidUnit.System.ENGLISH_LINEAR: "EnglishLinear",
                HidUnit.System.ENGLISH_ROTATION: "EnglishRotation",
            }

        def __str__(self: "HidUnit.System") -> str:
            return self._stringmap()[self]

        @classmethod
        def from_string(
            cls: _Type["HidUnit.System"], string: str
        ) -> Optional["HidUnit.System"]:
            """
            Returns the correct :class:`HidUnit.System` given the string.
            """
            try:
                return {v: k for k, v in cls._stringmap().items()}[string]
            except KeyError:
                return None

        @property
        def length(self: "HidUnit.System") -> Optional["Unit"]:
            """
            Returns the right :class:`Unit` for the length measurement in
            this system.
            """
            return {
                HidUnit.System.NONE: None,
                HidUnit.System.SI_LINEAR: Unit.CENTIMETER,
                HidUnit.System.SI_ROTATION: Unit.RADIANS,
                HidUnit.System.ENGLISH_LINEAR: Unit.INCH,
                HidUnit.System.ENGLISH_ROTATION: Unit.DEGREES,
            }[self]

        @property
        def mass(self: "HidUnit.System") -> Optional["Unit"]:
            """
            Returns the right :class:`Unit` for the mass measurement in
            this system.
            """
            return {
                HidUnit.System.NONE: None,
                HidUnit.System.SI_LINEAR: Unit.GRAM,
                HidUnit.System.SI_ROTATION: Unit.GRAM,
                HidUnit.System.ENGLISH_LINEAR: Unit.SLUG,
                HidUnit.System.ENGLISH_ROTATION: Unit.SLUG,
            }[self]

        @property
        def time(self: "HidUnit.System") -> Optional["Unit"]:
            """
            Returns the right :class:`Unit` for the time measurement in
            this system.
            """
            return {
                HidUnit.System.NONE: None,
                HidUnit.System.SI_LINEAR: Unit.SECONDS,
                HidUnit.System.SI_ROTATION: Unit.SECONDS,
                HidUnit.System.ENGLISH_LINEAR: Unit.SECONDS,
                HidUnit.System.ENGLISH_ROTATION: Unit.SECONDS,
            }[self]

        @property
        def temperature(self: "HidUnit.System") -> Optional["Unit"]:
            """
            Returns the right :class:`Unit` for the temperature measurement
            in this system.
            """
            return {
                HidUnit.System.NONE: None,
                HidUnit.System.SI_LINEAR: Unit.KELVIN,
                HidUnit.System.SI_ROTATION: Unit.KELVIN,
                HidUnit.System.ENGLISH_LINEAR: Unit.FAHRENHEIT,
                HidUnit.System.ENGLISH_ROTATION: Unit.FAHRENHEIT,
            }[self]

        @property
        def current(self: "HidUnit.System") -> Optional["Unit"]:
            """
            Returns the right :class:`Unit` for the current measurement
            in this system.
            """
            return {
                HidUnit.System.NONE: None,
                HidUnit.System.SI_LINEAR: Unit.AMPERE,
                HidUnit.System.SI_ROTATION: Unit.AMPERE,
                HidUnit.System.ENGLISH_LINEAR: Unit.AMPERE,
                HidUnit.System.ENGLISH_ROTATION: Unit.AMPERE,
            }[self]

        @property
        def luminous_intensity(self: "HidUnit.System") -> Optional["Unit"]:
            """
            Returns the right :class:`Unit` for the luminous intensity
            measurement in this system.
            """
            return {
                HidUnit.System.NONE: None,
                HidUnit.System.SI_LINEAR: Unit.CANDELA,
                HidUnit.System.SI_ROTATION: Unit.CANDELA,
                HidUnit.System.ENGLISH_LINEAR: Unit.CANDELA,
                HidUnit.System.ENGLISH_ROTATION: Unit.CANDELA,
            }[self]

    def __init__(
        self: "HidUnit", system: "HidUnit.System", units: Dict[Optional["Unit"], U16]
    ) -> None:
        self.units = units
        self.system = system

    @classmethod
    def _parse(cls: _Type["HidUnit"], data: Bytes) -> "HidUnit":
        assert data and len(data) >= 1

        def nibbles(data):
            for element in data:
                yield element & 0xF
                yield (element >> 4) & 0xF

        systems = (
            "System",
            "Length",
            "Mass",
            "Time",
            "Temperature",
            "Current",
            "Intensity",
            "Reserved",
        )

        # Creates a dict with the type of system as key and the value of the
        # nibble (the exponent) as value.
        exponents = dict(itertools.zip_longest(systems, nibbles(data)))
        system = HidUnit.System(exponents["System"])
        if system == HidUnit.System.NONE:
            return HidUnit.NONE

        def convert(exponent: Optional[U16]) -> Optional[U16]:
            return twos_comp(exponent, 4) if exponent is not None else None

        # Now create the mapping of correct unit types with their exponents, e.g.
        # {CENTIMETER: 2, SECONDS: -1}.
        units = {
            # system: convert(exponents['System']),
            system.length: convert(exponents["Length"]),
            system.mass: convert(exponents["Mass"]),
            system.time: convert(exponents["Time"]),
            system.temperature: convert(exponents["Temperature"]),
            system.current: convert(exponents["Current"]),
            system.luminous_intensity: convert(exponents["Intensity"]),
        }

        # Filter out any parts that aren't set
        units = {k: v for k, v in units.items() if v is not None and v}
        if units:
            return HidUnit(system, units)  # type: ignore ### bug in mypy it detects v from above as U16 | None
        else:
            return HidUnit.NONE

    @classmethod
    def from_bytes(cls: _Type["HidUnit"], data: Bytes) -> "HidUnit":
        """
        Converts the given data bytes into a :class:`HidUnit` object.
        The data bytes must not include the 0b011001nn prefix byte.

        Where the HID unit system is None, the returned value is None.
        """
        assert 1 <= len(data) <= 4
        return HidUnit._parse(data)

    @classmethod
    def from_value(cls: _Type["HidUnit"], value: Union[U8, U16, U32]) -> "HidUnit":
        """
        Converts the given 8, 16 or 32-bit value into a :class:`HidUnit`
        object.

        Where the HID unit system is None, the returned value is None.
        """
        assert value <= 0xFFFFFFFF
        return cls.from_bytes(value.to_bytes(byteorder="little", length=4))

    def __eq__(self: "HidUnit", other: Any) -> bool:
        if not isinstance(other, HidUnit):
            raise NotImplementedError
        return self.system == other.system and self.units == other.units

    def __str__(self: "HidUnit") -> str:
        units = []
        for u, exp in self.units.items():
            if exp == 1:
                s = ""
            else:
                s = "".join([superscripts[c] for c in str(exp)])
            if u is not None:
                units.append((u.value, s))

        # python 3.6 seems to not use __str__() for enums, leading to errors
        # in the test suite
        return f"{str(self.system)}: " + " * ".join(
            (f"{unit}{exp}" for unit, exp in units)
        )

    @classmethod
    def from_string(cls: _Type["HidUnit"], string: str) -> "HidUnit":
        """
        The inverse of __str__()
        """
        system_string, unit_string = string.split(": ")
        system = HidUnit.System.from_string(system_string)
        if system is None:
            return HidUnit.NONE

        unitstrings = unit_string.split(" * ")
        units: Dict[Optional["Unit"], U16] = {}
        for s in unitstrings:
            match: Optional[re.Match[str]]
            match = re.match(r"(?P<unit>[a-zA-z]+)(?P<exp>[⁰¹²³⁴⁵⁶⁷⁸⁹⁻]{1,})?", s)
            if match is None:
                continue
            unitstring, expstring = match["unit"], match["exp"]

            unit = Unit(unitstring)
            ssinv = {v: k for k, v in superscripts.items()}
            if expstring:
                exponent = int("".join([ssinv[c] for c in expstring]))
            else:
                exponent = 1

            units[unit] = exponent
        return HidUnit(system, units)

    @property
    def value(self: "HidUnit") -> U32:
        """
        Returns the numerical value for this unit as required by the HID
        specification.
        """
        v = self.system.value

        def unit_value(unit_type):
            if unit_type in self.units:
                return to_twos_comp(self.units[unit_type], 4)
            return 0

        v |= unit_value(self.system.length) << 4
        v |= unit_value(self.system.mass) << 8
        v |= unit_value(self.system.time) << 12
        v |= unit_value(self.system.temperature) << 16
        v |= unit_value(self.system.current) << 20
        v |= unit_value(self.system.luminous_intensity) << 24
        return v


class HidCollection:
    class Type(enum.IntEnum):
        PHYSICAL = 0
        APPLICATION = 1
        LOGICAL = 2
        REPORT = 3
        NAMED_ARRAY = 4
        USAGE_SWITCH = 5
        USAGE_MODIFIER = 6

    def __init__(self: "HidCollection", value: U8) -> None:
        assert value <= 0xFF
        self.value = value
        self.name = str(self)
        self.type: Optional[HidCollection.Type]
        try:
            self.type = HidCollection.Type(value)
        except ValueError:
            self.type = None

    @property
    def is_reserved(self: "HidCollection") -> bool:
        return 0x07 <= self.value <= 0x7F

    @property
    def is_vendor_defined(self: "HidCollection") -> bool:
        return 0x80 <= self.value <= 0xFF

    @classmethod
    def from_str(cls: _Type["HidCollection"], string: str) -> U16:
        """
        Return the value of this HidCollection given the human-readable
        string
        """
        for v in HidCollection.Type:
            if v.name == string.strip().upper():
                return v.value
        match = re.match(
            r"\s*(vendor[ -_]defined)\s+(0x|x)?(?P<value>[0-9a-f]{2,})",
            string,
            flags=re.IGNORECASE,
        )
        if not match:
            raise ValueError(f"Invalid HidCollection: {string}")

        return int(match["value"], 16)

    def __str__(self: "HidCollection") -> str:
        try:
            return HidCollection.Type(self.value).name
        except ValueError:
            if self.is_reserved:
                c = f"RESERVED {self.value:#x}"
            elif self.is_vendor_defined:
                c = f"VENDOR_DEFINED {self.value:#x}"
            else:  # not supposed to happen
                raise
            return c


sensor_mods: Final = {
    0x00: "Mod None",
    0x10: "Mod Change Sensitivity Abs",
    0x20: "Mod Max",
    0x30: "Mod Min",
    0x40: "Mod Accuracy",
    0x50: "Mod Resolution",
    0x60: "Mod Threshold High",
    0x70: "Mod Threshold Low",
    0x80: "Mod Calibration Offset",
    0x90: "Mod Calibration Multiplier",
    0xA0: "Mod Report Interval",
    0xB0: "Mod Frequency Max",
    0xC0: "Mod Period Max",
    0xD0: "Mod Change Sensitivity Range Percent",
    0xE0: "Mod Change Sensitivity Rel Percent",
    0xF0: "Mod Vendor Reserved",
}

inv_hid: Dict[U16, str] = {}  # e.g 0b10000000 : "Input"
hid_type: Dict[str, str] = {}  # e.g. "Input" : "Main"
for type, items in hid_items.items():
    for k, v in items.items():
        inv_hid[v] = k
        hid_type[k] = type


class ParseError(Exception):
    """Exception thrown during report descriptor parsing"""

    pass


class RangeError(Exception):
    """Exception thrown for an out-of-range value

    .. attribute:: value

        The invalid value

    .. attribute:: range

        A ``(min, max)`` tuple for the allowed logical range

    .. attribute:: field

        The :class:`HidField` that triggered this exception
    """

    def __init__(self: "RangeError", field: "HidField", value: int) -> None:
        self.field = field
        self.range = field.logical_min, field.logical_max
        self.value = value

    def __str__(self: "RangeError") -> str:
        min, max = self.range
        return f"Value {self.value} is outside range {min}, {max} for {self.field.usage_name}"


class Unit(enum.Enum):
    CENTIMETER = "cm"
    RADIANS = "rad"
    INCH = "in"
    DEGREES = "deg"
    GRAM = "g"
    SLUG = "slug"
    SECONDS = "s"
    KELVIN = "K"
    FAHRENHEIT = "F"
    AMPERE = "A"
    CANDELA = "cd"
    RESERVED = "reserved"


class HidField(object):
    """
    Represents one field in a HID report. A field is one element of a HID
    report that matches a specific set of bits in that report.

    .. attribute:: usage

        The numerical HID field's Usage, e.g. 0x38 for "Wheel". If the field
        has multiple usages, this refers to the first one.

    .. attribute:: usage_page

        The numerical HID field's Usage Page, e.g. 0x01 for "Generic
        Desktop"

    .. attribute:: report_ID

        The numeric Report ID this HID field belongs to

    .. attribute:: logical_min

        The logical minimum of this HID field

    .. attribute:: logical_max

        The logical maximum of this HID field

    .. attribute:: physical_min

        The physical minimum of this HID field

    .. attribute:: physical_max

        The physical maximum of this HID field

    .. attribute:: unit

        The unit of this HID field

    .. attribute:: unit_exp

        The unit exponent of this HID field

    .. attribute:: size

        Report Size in bits for this HID field

    .. attribute:: count

        Report Count for this HID field
    """

    def __init__(
        self: "HidField",
        report_ID: U8,
        logical: Optional[U32],
        physical: Optional[U32],
        application: Optional[U32],
        collection: Optional[Tuple[U32, U32, U32]],
        value: U32,
        usage_page: U16,
        usage: U32,
        logical_min: U32,
        logical_max: U32,
        physical_min: U32,
        physical_max: U32,
        unit: U16,
        unit_exp: U8,
        item_size: U8,
        count: U8,
    ) -> None:
        self.report_ID = report_ID
        self.logical = logical
        self.physical = physical
        self.application = application
        self.collection = collection
        self.type = value
        self.usage_page = usage_page
        self.usage = usage
        self.usages: Optional[List[U32]] = None
        self.logical_min = logical_min
        self.logical_max = logical_max
        self.physical_min = physical_min
        self.physical_max = physical_max
        self.unit = unit
        self.unit_exp = unit_exp
        self.size = item_size
        self.count = count
        self.start = 0

    def copy(self: "HidField") -> "HidField":
        """
        Return a full copy of this :class:`HIDField`.
        """
        c = copy.copy(self)
        if self.usages is not None:
            c.usages = self.usages[:]
        return c

    def _usage_name(self: "HidField", usage: U32) -> str:
        usage_page: U16 = usage >> 16
        value: U16 = usage & 0x0000FFFF
        if usage_page in HUT:  # type: ignore ### Operator "in" not supported for types "int" and "HidUsageTable"
            if HUT[usage_page].page_name == "Button":
                name = f"B{str(value)}"
            else:
                try:
                    name = HUT[usage_page][value].name
                except KeyError:
                    name = f"0x{usage:04x}"
        else:
            name = f"0x{usage:04x}"
        return name

    @property
    def usage_name(self: "HidField") -> str:
        """
        The Usage name for this field (e.g. "Wheel").
        """
        return self._usage_name(self.usage)

    def get_usage_name(self: "HidField", index: int) -> Optional[str]:
        """
        Return the Usage name for this field at the given index. Use this
        function when the HID field has multiple Usages.
        """
        if self.usages is not None:
            return self._usage_name(self.usages[index])
        return None

    @property
    def physical_name(self: "HidField") -> Optional[str]:
        """
        The physical name or ``None``
        """
        phys = self.physical
        if phys is None:
            return phys

        _phys = ""
        try:
            page_id = phys >> 16
            value = phys & 0xFF
            _phys = HUT[page_id][value].name
        except KeyError:
            try:
                _phys = f"0x{phys:04x}"
            except ValueError:
                pass
        return _phys

    @property
    def logical_name(self: "HidField") -> Optional[str]:
        """
        The logical name or ``None``
        """
        logical = self.logical
        if logical is None:
            return None

        _logical = ""

        try:
            page_id = logical >> 16
            value = logical & 0xFF
            _logical = HUT[page_id][value].name
        except KeyError:
            try:
                _logical = f"0x{logical:04x}"
            except ValueError:
                pass
        return _logical

    def _get_value(self: "HidField", report: List[U8], idx: int) -> Union[U32, str]:
        """
        Extract the bits that are this HID field in the list of bytes
        ``report``

        :param list report: a list of bytes that represent a HID report
        :param int idx: which field index to fetch, only greater than 0 if
            :attr:`count` is larger than 1
        """
        value = 0
        start_bit = self.start + self.size * idx
        end_bit = start_bit + self.size * (idx + 1)
        data = report[int(start_bit / 8) : int(end_bit / 8 + 1)]
        if len(data) == 0:
            return "<.>"
        for d in range(len(data)):
            value |= data[d] << (8 * d)

        bit_offset = start_bit % 8
        value = value >> bit_offset
        garbage = (value >> self.size) << self.size
        value = value - garbage
        if self.logical_min < 0 and self.size > 1:
            value = twos_comp(value, self.size)
        return value

    def get_values(self: "HidField", report: List[U8]) -> List[Union[U32, str]]:
        """
        Assume ``report`` is a list of bytes that are a full HID report,
        extract the values that are this HID field.

        Example:

        - if this field is Usage ``X`` , this returns ``[x-value]``
        - if this field is Usage ``X``, ``Y``, this returns ``[x, y]``
        - if this field is a button mask, this returns ``[1, 0, 1, ...]``, i.e. one value for each
          button

        :param list report: a list of bytes that are a HID report
        :returns: a list of integer values of len :attr:`count`
        """
        return [self._get_value(report, i) for i in range(self.count)]

    def _fill_value(self: "HidField", report: List[U8], value: U32, idx: int) -> None:
        start_bit = self.start + self.size * idx
        n = self.size

        max = (1 << n) - 1
        if value > max:
            raise Exception(
                f"_set_value(): value {value} is larger than size {self.size}"
            )

        byte_idx = int(start_bit / 8)
        bit_shift = start_bit % 8
        bits_to_set = 8 - bit_shift

        while n - bits_to_set >= 0:
            report[byte_idx] &= ~(0xFF << bit_shift)
            report[byte_idx] |= (value << bit_shift) & 0xFF
            value >>= bits_to_set
            n -= bits_to_set
            bits_to_set = 8
            bit_shift = 0
            byte_idx += 1

        # last nibble
        if n:
            bit_mask = (1 << n) - 1
            report[byte_idx] &= ~(bit_mask << bit_shift)
            report[byte_idx] |= value << bit_shift

    def fill_values_array(self: "HidField", report: List[U8], data: List[Any]) -> None:
        """
        Assuming ``data`` is the value for this HID field array and ``report``
        is a HID report's bytes, this method sets those bits in ``report`` that
        are his HID field to ``value``.

        Example:
        - if this field is an array of keys, use
          ``fill_values(report, ['a or A', 'b or B', ...]``, i.e. one string
          representation for each pressed key


        ``data`` must have at most :attr:`count` elements, matching this
        field's Report Count.


        :param list report: an integer array representing this report,
            modified in place
        :param list data: the data for this hid field with one element for
            each Usage.
        """
        if len(data) > self.count:
            raise Exception("-EINVAL")

        array: List[int] = []

        for usage_name in data:
            try:
                usage = HUT[self.usage_page].from_name[usage_name]
            except KeyError:
                continue

            full_usage = usage.usage_page.page_id << 16 | usage.usage

            if self.usages is not None and full_usage in self.usages:
                idx = self.usages.index(full_usage)
                array.append(idx)

        for idx in range(self.count):
            try:
                v = array[idx]
            except IndexError:
                v = 0

            v += self.logical_min

            self._fill_value(report, v, idx)

    def fill_values(self: "HidField", report: List[U8], data: List[U32]) -> None:
        """
        Assuming ``data`` is the value for this HID field and ``report`` is
        a HID report's bytes, this method sets those bits in ``report`` that
        are his HID field to ``value``.

        Example:

        - if this field is Usage ``X`` , use ``fill_values(report, [x-value])``
        - if this field is Usage ``X``, ``Y``, use ``fill_values(report, [x, y])``
        - if this field is a button mask, use
          ``fill_values(report, [1, 0, 1, ...]``, i.e. one value for each
          button

        ``data`` must have at least :attr:`count` elements, matching this
        field's Report Count.


        :param list report: an integer array representing this report,
            modified in place
        :param list data: the data for this hid field with one element for
            each Usage.
        """
        if len(data) != self.count:
            raise Exception("-EINVAL")

        for idx in range(self.count):
            v = data[idx]

            if self.is_null:
                # FIXME: handle the signed case too
                if v >= (1 << self.size):
                    raise RangeError(self, v)
            elif self.usage_name not in ["Contact Id", "Contact Max", "Contact Count"]:
                if v and not (self.logical_min <= v <= self.logical_max):
                    raise RangeError(self, v)
            if self.logical_min < 0:
                v = to_twos_comp(v, self.size)
            self._fill_value(report, v, idx)

    @property
    def is_array(self: "HidField") -> bool:
        """
        ``True`` if this HID field is an array
        """
        return not (self.type & (0x1 << 1))  # Variable

    @property
    def is_const(self: "HidField") -> bool:
        """
        ``True`` if this HID field is const
        """
        return bool(self.type & (0x1 << 0))

    @property
    def is_null(self: "HidField") -> bool:
        """
        ``True`` if this HID field is null
        """
        return bool(self.type & (0x1 << 6))

    @property
    def usage_page_name(self: "HidField") -> str:
        """
        The Usage Page name for this field, e.g. "Generic Desktop"
        """
        usage_page_name = ""
        usage_page = self.usage_page >> 16
        try:
            usage_page_name = HUT[usage_page].page_name
        except KeyError:
            pass
        return usage_page_name

    @classmethod
    def getHidFields(
        cls: _Type["HidField"],
        report_ID: U8,
        logical: Optional[U32],
        physical: Optional[U32],
        application: Optional[U32],
        collection: Optional[Tuple[U32, U32, U32]],
        value: U32,
        usage_page: U16,
        usages: List[U32],
        usage_min: U32,
        usage_max: U32,
        logical_min: U32,
        logical_max: U32,
        physical_min: U32,
        physical_max: U32,
        unit: U16,
        unit_exp: U8,
        item_size: U8,
        count: int,
    ):
        """
        This is a function to be called by a HID report descriptor parser.

        Given the current parser state and the various arguments, create the
        required number of :class:`HidField` objects.

        :returns: a list of :class:`HidField` objects
        """

        # Note: usage_page is a 32-bit value here and usage
        # is usage_page | usage

        usage: U32 = usage_min
        if len(usages) > 0:
            usage = usages[0]

        # for arrays, we don't have a given usage
        # use either the logical if given or the application
        if not value & 0x3:
            if logical is not None and logical:
                usage = logical
            elif application is not None:
                usage = application

        item = cls(
            report_ID,
            logical,
            physical,
            application,
            collection,
            value,
            usage_page,
            usage,
            logical_min,
            logical_max,
            physical_min,
            physical_max,
            unit,
            unit_exp,
            item_size,
            1,
        )
        items = []

        if value & 0x3:  # Const or Variable item
            if usage_min and usage_max:
                usage = usage_min
                for i in range(count):
                    item = item.copy()
                    item.usage = usage
                    items.append(item)
                    if usage < usage_max:
                        usage += 1
            elif usages:
                for i in range(count):
                    if i < len(usages):
                        usage = usages[i]
                    else:
                        usage = usages[-1]
                    item = item.copy()
                    item.usage = usage
                    items.append(item)
            # A const field used for padding may not have any usages
            else:
                item.size *= count
                return [item]
        else:  # Array item
            if usage_min and usage_max:
                usages = list(range(usage_min, usage_max + 1))
            item.usages = usages
            item.count = count
            return [item]
        return items


class HidReport(object):
    """
    Represents a HidReport, one of ``Input``, ``Output``, ``Feature``. A
    :class:`ReportDescriptor` may contain one or more
    HidReports of different types. These comprise of a number of
    :class:`HidField` making up the exact description of a
    report.

    :param int report_ID: the report ID
    :param int application: the HID application

    .. attribute:: fields

        The :class:`HidField` elements comprising this report

    """

    class Type(enum.Enum):
        """The type of a :class:`HidReport`"""

        INPUT = enum.auto()
        OUTPUT = enum.auto()
        FEATURE = enum.auto()

    def __init__(
        self: "HidReport",
        report_ID: U8,
        application: Optional[U32],
        type: "HidReport.Type",
    ) -> None:
        self.fields: List[HidField] = []
        self.report_ID = report_ID
        self.application = application
        self._application_name: Optional[str] = None
        self._bitsize = 0
        if self.numbered:
            self._bitsize = 8
        self._type = type
        self.prev_collection: Optional[Tuple[U32, U32, U32]] = None

    def append(self: "HidReport", field: HidField) -> None:
        """
        Add a :class:`HidField` to this report

        :param HidField field: the object to add to this report
        """
        self.fields.append(field)
        field.start = self._bitsize
        self._bitsize += field.size

    def extend(self: "HidReport", fields: List[HidField]) -> None:
        """
        Extend this report by the list of :class:`HidField`
        objects

        :param list fields: a list of objects to append to this report
        """
        self.fields.extend(fields)
        for f in fields:
            f.start = self._bitsize
            self._bitsize += f.size * f.count

    @property
    def application_name(self: "HidReport") -> str:
        if self.application is None:
            return "Vendor"

        try:
            page_id = self.application >> 16
            value = self.application & 0xFF
            return HUT[page_id][value].name
        except KeyError:
            return "Vendor"

    @property
    def type(self: "HidReport") -> "HidReport.Type":
        """
        One of the types in :class:`HidReport.Type`
        """
        return self._type

    @property
    def numbered(self: "HidReport") -> bool:
        """
        True if the HidReport was initialized with a report ID
        """
        return self.report_ID >= 0

    @property
    def bitsize(self: "HidReport") -> int:
        """
        The size of the HidReport in bits
        """
        return self._bitsize

    @property
    def size(self: "HidReport") -> int:
        """
        The size of the HidReport in bytes
        """
        return self._bitsize >> 3


class _HidRDescItem(object):
    """
    Represents one item in the Report Descriptor. This is a variable-sized
    element with one header byte and 0, 1, 2, 4 payload bytes.

    :param int index_in_report:
        The index within the report descriptor
    :param int hid:
        The numerical hid type (e.g. ``0b00000100`` for Usage Page)
    :param int value:
        The 8, 16, or 32 bit value
    :param list raw_values:
        The payload bytes' raw values, LSB first


    These items are usually parsed from a report descriptor, see
    :meth:`from_bytes`. The report descriptor
    bytes are::

                H P P H H P H P

    where each header byte looks like this

    +---------+---+---+---+---+---+---+---+---+
    | bit     | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
    +=========+===+===+===+===+===+===+===+===+
    |         |   hid item type       | size  |
    +---------+-----------------------+-------+

    .. note:: a size of 0x3 means payload size 4

    To create a _HidRDescItem from a human-readable description, use
    :meth:`from_human_descr`.



    .. attribute:: index_in_report

        The numerical index of this item in the report descriptor.

    .. attribute:: raw_value

        A list of the payload's raw values

    .. attribute:: hid

        The hid item as number (e.g. ``0b00000100`` for Usage Page)

    .. attribute:: item

        The hid item as string (e.g. "Usage Page")

    .. attribute:: value

        The payload value as single number

    """

    def __init__(
        self: "_HidRDescItem",
        index_in_report: int,
        hid: U16,
        value: int,
        raw_values: List[U8],
    ) -> None:
        self.index_in_report = index_in_report
        self.raw_value = raw_values
        self.hid = hid
        self.value = value
        self.usage_page: U16 = 0
        try:
            self.item = inv_hid[self.hid]
        except KeyError:
            error = f"error while parsing {hid:02x}"
            raise KeyError(error)

        if self.item in (
            "Logical Minimum",
            "Physical Minimum",
            # "Logical Maximum",
            # "Physical Maximum",
        ):
            self._twos_comp()
        if self.item == "Unit Exponent" and self.value > 7:
            self.value -= 16

    def _twos_comp(self: "_HidRDescItem") -> int:
        self.value = twos_comp(self.value, (self.size - 1) * 8)
        return self.value

    @property
    def size(self: "_HidRDescItem") -> int:
        """The size in bytes, including header byte"""
        return 1 + len(self.raw_value)

    @property
    def bytes(self: "_HidRDescItem") -> List[U8]:
        """
        Return this in the original format in bytes, i.e. a header byte
        followed by (if any) payload bytes.

        :returns: a list of bytes that are this item
        """
        if len(self.raw_value) == 4:
            h = self.hid | 0x3
        else:
            h = self.hid | len(self.raw_value)
        return [h] + self.raw_value.copy()

    def __repr__(self: "_HidRDescItem") -> str:
        data = [f"{i:02x}" for i in self.bytes]
        return f'{" ".join(data)}'

    def _get_raw_values(self: "_HidRDescItem") -> str:
        """The raw values as comma-separated hex numbers"""
        data = str(self)
        # prefix each individual value by "0x" and insert "," in between
        data = f'0x{data.replace(" ", ", 0x")},'
        return data

    def get_human_descr(self: "_HidRDescItem", indent: int) -> Tuple[str, int]:
        """
        Return a human-readable description of this item

        :param int indent: The indentation to prefix
        """
        item = self.item
        value = self.value
        up = self.usage_page
        descr = item
        # Use a prefix to signify attrs that apply to the next input/output
        prefix = " ."
        if item in (
            "Report ID",
            "Usage Minimum",
            "Usage Maximum",
            "Logical Minimum",
            "Physical Minimum",
            "Logical Maximum",
            "Physical Maximum",
            "Report Size",
            "Report Count",
            "Unit Exponent",
        ):
            if item != "Report ID":
                descr = prefix + descr
            descr += f" ({str(value)})"
        elif item == "Collection":
            descr += f" ({HidCollection(value).name.capitalize()})"
            indent += 1
        elif item == "End Collection":
            indent -= 1
        elif item == "Usage Page":
            try:
                descr += f" ({HUT[value].page_name})"
            except KeyError:
                descr += f" (Vendor Usage Page 0x{value:02x})"
        elif item == "Usage":
            usage = value | up
            try:
                descr += f" ({HUT[up >> 16][value]})"
            except KeyError:
                sensor = HUT.usage_page_from_name("Sensor")
                if sensor is not None and (up >> 16) == sensor.page_id:
                    mod = (usage & 0xF000) >> 8
                    usage &= ~0xF000
                    mod_descr = sensor_mods[mod]
                    page_id = (usage & 0xFF00) >> 16
                    try:
                        descr += f" ({HUT[page_id][usage & 0xFF]}  | {mod_descr})"
                    except KeyError:
                        descr += f" (Unknown Usage 0x{value:02x})"
                else:
                    descr += f" (Vendor Usage 0x{value:02x})"
        elif item == "Input" or item == "Output" or item == "Feature":
            descr += " ("
            if value & (0x1 << 0):
                descr += "Cnst,"
            else:
                descr += "Data,"
            if value & (0x1 << 1):
                descr += "Var,"
            else:
                descr += "Arr,"
            if value & (0x1 << 2):
                descr += "Rel"
            else:
                descr += "Abs"
            if value & (0x1 << 3):
                descr += ",Wrap"
            if value & (0x1 << 4):
                descr += ",NonLin"
            if value & (0x1 << 5):
                descr += ",NoPref"
            if value & (0x1 << 6):
                descr += ",Null"
            if value & (0x1 << 7):
                descr += ",Vol"
            if value & (0x1 << 8):
                descr += ",Buff"
            descr += ")"
        elif item == "Unit":
            unit = HidUnit.from_value(value)
            descr += f" ({unit})"
        elif item == "Push":
            pass
        elif item == "Pop":
            pass
        eff_indent = indent
        if item == "Collection":
            eff_indent -= 1
        return " " * eff_indent + descr, indent

    @classmethod
    def _one_item_from_bytes(
        cls: _Type["_HidRDescItem"], rdesc: Union[Bytes, List[U8]]
    ) -> Optional["_HidRDescItem"]:
        """
        Parses a single (the first) item from the given report descriptor.

        :param rdesc: a series of bytes representing the report descriptor

        :returns: a single _HidRDescItem from the first ``item.size`` bytes
                of the descriptor

        .. note:: ``item.index_in_report`` is always 0 when using this function
        """
        idx = 0
        header = rdesc[idx]
        if header == 0 and idx == len(rdesc) - 1:
            # some devices present a trailing 0, skipping it
            return None

        index_in_report = 0  # always zero, oh well
        size = header & 0x3
        if size == 3:
            size = 4
        hid = header & 0xFC
        if hid == 0:
            raise ParseError(f"Unexpected HID type 0 in {header:02x}")

        value = 0
        raw_values = []

        idx += 1
        if size >= 1:
            v = rdesc[idx]
            idx += 1
            raw_values.append(v)
            value |= v
        if size >= 2:
            v = rdesc[idx]
            idx += 1
            raw_values.append(v)
            value |= v << 8
        if size >= 4:
            v = rdesc[idx]
            idx += 1
            raw_values.append(v)
            value |= v << 16
            v = rdesc[idx]
            idx += 1
            raw_values.append(v)
            value |= v << 24

        return _HidRDescItem(index_in_report, hid, value, raw_values)

    @classmethod
    def from_bytes(
        cls: _Type["_HidRDescItem"],
        rdesc: Union[
            Bytes,
            List[U8],
        ],
    ) -> List["_HidRDescItem"]:
        """
        Parses a series of bytes into items.

        :param list rdesc: a series of bytes that are a HID report
                descriptor

        :returns: a list of items representing this report descriptor
        """
        items = []
        idx = 0
        while idx < len(rdesc):
            item = _HidRDescItem._one_item_from_bytes(rdesc[idx:])
            if item is None:
                break
            item.index_in_report = idx
            items.append(item)
            idx += item.size

        return items

    @classmethod
    def from_human_descr(
        cls: _Type["_HidRDescItem"], line: str, usage_page: U16
    ) -> "_HidRDescItem":
        """
        Parses a line from human-readable HID report descriptor e.g.::

            Usage Page (Digitizers)
            Usage (Finger)
            Collection (Logical)
             Report Size (1)
             Report Count (1)
             Logical Minimum (0)
             Logical Maximum (1)
             Usage (Tip Switch)
             Input (Data,Var,Abs)


        :param str line: a single line in the report descriptor
        :param int usage_page: the usage page to set for this item

        :returns: a single item representing the current line
        """
        data = None
        if "(" in line:
            m = re.match(r"\s*(?P<name>[^(]+)\((?P<data>.+)\)", line)
            assert m is not None
            name = m.group("name").strip()
            data = m.group("data")
            if data.lower().startswith("0x"):
                try:
                    data = int(data[2:], 16)
                except ValueError:
                    pass
            else:
                try:
                    data = int(data)
                except ValueError:
                    pass
        # Strip any superfluous stuff from an EndCollection line
        elif "End Collection" in line:
            name = "End Collection"
        else:
            name = line.strip()

        value = None

        def hex_value(string: str, prefix: str) -> Optional[U16]:
            if string.startswith(prefix):
                return int(string[len(prefix) :], 16)
            return None

        if isinstance(data, str):
            if name == "Usage Page":
                up = HUT.usage_page_from_name(data)
                if up is None:
                    prefix = "Vendor Usage Page "
                    assert data.startswith(prefix)
                    value = hex_value(data, prefix)
                else:
                    page = HUT.usage_page_from_name(data)
                    if page is not None:
                        value = page.page_id
                if value is not None:
                    usage_page = value
            elif name == "Usage":
                try:
                    value = HUT[usage_page].from_name[data].usage
                except KeyError:
                    value = hex_value(data, "Vendor Usage ")
                    if value is None:
                        raise
            elif name == "Collection":
                value = HidCollection.from_str(data)
            elif name in "Input Output Feature":
                value = 0
                possible_types = (
                    "Cnst",
                    "Var",
                    "Rel",
                    "Wrap",
                    "NonLin",
                    "NoPref",
                    "Null",
                    "Vol",
                    "Buff",
                )
                for i, t in enumerate(possible_types):
                    if t in data:
                        value |= 0x1 << i
            elif name == "Unit":
                if data == "None":
                    value = 0
                else:
                    value = HidUnit.from_string(data).value
        else:  # data has been converted to an int already
            if name == "Usage Page" and data is not None:
                usage_page = data
            value = data

        v_count = 0
        if value is not None:
            if value <= 0xFF:
                v_count = 1
            elif value <= 0xFFFF:
                v_count = 2
            else:
                v_count = 4
        else:
            value = 0
        tag = hid_items[hid_type[name]][name]

        if value < 0:
            if name == "Unit Exponent":
                value += 16
                value = to_twos_comp(value, v_count * 8)
            elif name in ("Logical Minimum", "Physical Minimum"):
                value = to_twos_comp(value, v_count * 8)

        assert value is not None

        v: U16 = value
        vs = []
        for i in range(v_count):
            vs.append(v & 0xFF)
            v >>= 8

        item = _HidRDescItem(0, tag, value, vs)
        item.usage_page = usage_page << 16

        return item

    def dump_rdesc_kernel(self: "_HidRDescItem", indent: int, dump_file: IO) -> int:
        """
        Write the HID item to the file a C-style format, e.g. ::

            0x05, 0x01,			/* Usage Page (Generic Desktop)			*/

        :param int indent: indentation to prefix
        :param File dump_file: file to write to
        """
        # offset = self.index_in_report
        line = self._get_raw_values()
        line += "\t" * (int((40 - len(line)) / 8))

        descr, indent = self.get_human_descr(indent)

        descr += "\t" * (int((52 - len(descr)) / 8))
        # dump_file.write(f'{line}/* {descr} {str(offset)} */\n')
        dump_file.write(f"\t{line}/* {descr}*/\n")
        return indent

    def dump_rdesc_array(self: "_HidRDescItem", indent: int, dump_file: IO) -> int:
        """
        Format the hid item in hexadecimal format with a
        double-slash comment, e.g. ::

           0x05, 0x01,                    // Usage Page (Generic Desktop)        0

        :param int indent: indentation to prefix
        :param File dump_file: file to write to
        """
        offset = self.index_in_report
        line = self._get_raw_values()

        descr, indent = self.get_human_descr(indent)

        dump_file.write(f"{line:18s} // {offset:03x}:  {descr}\n")
        return indent

    def dump_rdesc_human(self: "_HidRDescItem", indent: int, dump_file: IO) -> int:
        """
        Format the hid item in human-only format e.g. ::

           Usage Page (Generic Desktop)        0

        :param int indent: indentation to prefix
        :param File dump_file: file to write to
        """
        offset = self.index_in_report
        descr, indent = self.get_human_descr(indent)
        descr += " " * (35 - len(descr))
        dump_file.write(f"{descr} {str(offset)}\n")
        return indent


class ReportDescriptor(object):
    """
    Represents a fully parsed HID report descriptor.

    When creating a ``ReportDescriptor`` object,

    - if your source is a stream of bytes, use
      :meth:`from_bytes`
    - if your source is a human-readable descriptor, use
      :meth:`from_human_descr`

    .. attribute:: win8

        ``True`` if the device is Windows8 compatible, ``False`` otherwise

    .. attribute:: input_reports

        All :class:`HidReport` of type ``Input``, addressable by the report ID

    .. attribute:: output_reports

        All :class:`HidReport` of type ``Output``, addressable by the report ID

    .. attribute:: feature_reports

        All :class:`HidReport` of type ``Feature``, addressable by the report ID
    """

    class _Globals(object):
        """
        HID report descriptors uses a stack-based model where some values
        are pushed to the global state and apply to all subsequent items
        until changed or reset.
        """

        def __init__(
            self: "ReportDescriptor._Globals",
            other: Optional["ReportDescriptor._Globals"] = None,
        ) -> None:
            self.usage_page: U16 = 0
            self.logical: Optional[U32] = None
            self.physical: Optional[U32] = None
            self.application: Optional[U32] = None
            self.logical_min: U32 = 0
            self.logical_max: U32 = 0
            self.physical_min: U32 = 0
            self.physical_max: U32 = 0
            self.unit: U32 = 0
            self.unit_exp: U32 = 0
            self.count: int = 0
            self.item_size: int = 0
            if other is not None:
                self.usage_page = other.usage_page
                self.logical = other.logical
                self.physical = other.physical
                self.application = other.application
                self.logical_min = other.logical_min
                self.logical_max = other.logical_max
                self.physical_min = other.physical_min
                self.physical_max = other.physical_max
                self.unit = other.unit
                self.unit = other.unit_exp
                self.count = other.count
                self.item_size = other.item_size

    class _Locals(object):
        """
        HID report descriptors uses a stack-based model where values
        apply until the next Output/InputReport/FeatureReport item.
        """

        def __init__(self: "ReportDescriptor._Locals") -> None:
            self.usages: List[U32] = []
            self.usage_sizes: List[int] = []
            self.usage_min: U32 = 0
            self.usage_max: U32 = 0
            self.usage_max_size: U32 = 0
            self.report_ID: U8 = -1

    def __init__(self: "ReportDescriptor", items: List[_HidRDescItem]) -> None:
        self.input_reports: Dict[U8, HidReport] = {}
        self.feature_reports: Dict[U8, HidReport] = {}
        self.output_reports: Dict[U8, HidReport] = {}
        self.win8: bool = False
        self.rdesc_items = items

        # variables only used during parsing
        self.global_stack: List["ReportDescriptor._Globals"] = []
        self.collection: List[U32] = [0, 0, 0]  # application, physical, logical
        self.local = ReportDescriptor._Locals()
        self.glob: "ReportDescriptor._Globals" = ReportDescriptor._Globals()
        self.current_item = None

        index_in_report = 0
        for item in items:
            item.index_in_report = index_in_report
            index_in_report += item.size
            self._parse_item(item)

        # Drop the parsing-only variables so we don't leak them later
        del self.current_item
        del self.glob
        del self.global_stack
        del self.local
        del self.collection

    def get(
        self: "ReportDescriptor", reportID: U8, reportSize: int
    ) -> Optional[HidReport]:
        """
        Return the input report with the given Report ID or ``None``
        """
        try:
            report = self.input_reports[reportID]
        except KeyError:
            try:
                report = self.input_reports[-1]
            except KeyError:
                return None

        # if the report is larger than it should be, it's OK,
        # extra bytes will just be ignored
        if report.size <= reportSize:
            return report

        return None

    def get_report_from_application(
        self: "ReportDescriptor", application: Union[str, U32]
    ) -> Optional[HidReport]:
        """
        Return the Input report that matches the application or ``None``
        """
        for r in self.input_reports.values():
            if r.application == application or r.application_name == application:
                return r
        return None

    def _get_current_report(self: "ReportDescriptor", type: str) -> HidReport:
        report_lists = {
            "Input": self.input_reports,
            "Output": self.output_reports,
            "Feature": self.feature_reports,
        }
        report_type = {
            "Input": HidReport.Type.INPUT,
            "Output": HidReport.Type.OUTPUT,
            "Feature": HidReport.Type.FEATURE,
        }

        assert type in report_lists

        try:
            cur = report_lists[type][self.local.report_ID]
        except KeyError:
            cur = HidReport(
                self.local.report_ID, self.glob.application, report_type[type]  # type: ignore
            )
            report_lists[type][self.local.report_ID] = cur
        return cur

    def _concatenate_usages(self: "ReportDescriptor") -> None:
        if self.local.usage_max and self.local.usage_max_size <= 2:
            if self.local.usage_max & 0xFFFF0000 != self.glob.usage_page:
                self.local.usage_max &= 0xFFFF
                self.local.usage_max |= self.glob.usage_page
                self.local.usage_min &= 0xFFFF
                self.local.usage_min |= self.glob.usage_page

        for i, v in reversed(list(enumerate(self.local.usages))):
            if self.local.usage_sizes[i] > 2:
                continue
            if v & 0xFFFF0000 == self.glob.usage_page:
                break
            self.local.usages[i] = v & 0xFFFF | self.glob.usage_page

    def _parse_item(self: "ReportDescriptor", rdesc_item: _HidRDescItem) -> None:
        # store current usage_page in rdesc_item
        rdesc_item.usage_page = self.glob.usage_page
        item = rdesc_item.item
        value = rdesc_item.value
        size = rdesc_item.size - 1

        if item == "Report ID":
            self.local.report_ID = value
        elif item == "Push":
            self.global_stack.append(self.glob)
            self.glob = ReportDescriptor._Globals(self.glob)
        elif item == "Pop":
            self.glob = self.global_stack.pop()
        elif item == "Usage Page":
            self.glob.usage_page = value << 16
        elif item == "Collection":
            self._concatenate_usages()

            c = HidCollection(value)
            try:
                if c.type == HidCollection.Type.PHYSICAL:
                    self.collection[1] += 1
                    self.glob.physical = self.local.usages[-1]
                elif c.type == HidCollection.Type.APPLICATION:
                    self.collection[0] += 1
                    self.glob.application = self.local.usages[-1]
                elif c.type == HidCollection.Type.LOGICAL:
                    self.collection[2] += 1
                    self.glob.logical = self.local.usages[-1]
            except IndexError:
                pass
            # reset the usage list
            self.local.usages = []
            self.local.usage_sizes = []
            self.local.usage_min = 0
            self.local.usage_max = 0
            self.local.usage_max_size = 0
        elif item == "Usage Minimum":
            if size <= 2:
                self.local.usage_min = value | self.glob.usage_page
            else:
                self.local.usage_min = value
        elif item == "Usage Maximum":
            if size <= 2:
                self.local.usage_max = value | self.glob.usage_page
            else:
                self.local.usage_max = value
            self.local.usage_max_size = size
        elif item == "Logical Minimum":
            self.glob.logical_min = value
        elif item == "Logical Maximum":
            self.glob.logical_max = value
        elif item == "Physical Minimum":
            self.glob.physical_min = value
        elif item == "Physical Maximum":
            self.glob.physical_max = value
        elif item == "Unit":
            self.glob.unit = value
        elif item == "Unit Exponent":
            self.glob.unit_exp = value
        elif item == "Usage":
            if size <= 2:
                self.local.usages.append(value | self.glob.usage_page)
            else:
                self.local.usages.append(value)
            self.local.usage_sizes.append(size)
        elif item == "Report Count":
            self.glob.count = value
        elif item == "Report Size":
            self.glob.item_size = value
        elif item in ("Input", "Feature", "Output"):
            self.current_input_report = self._get_current_report(item)

            self._concatenate_usages()

            inputItems = HidField.getHidFields(
                self.local.report_ID,
                self.glob.logical,
                self.glob.physical,
                self.glob.application,
                cast(Tuple[U32, U32, U32], tuple(self.collection)),
                value,
                self.glob.usage_page,
                self.local.usages,
                self.local.usage_min,
                self.local.usage_max,
                self.glob.logical_min,
                self.glob.logical_max,
                self.glob.physical_min,
                self.glob.physical_max,
                self.glob.unit,
                self.glob.unit_exp,
                self.glob.item_size,
                self.glob.count,
            )
            self.current_input_report.extend(inputItems)
            if (
                item == "Feature"
                and len(self.local.usages) > 0
                and self.local.usages[-1] == 0xFF0000C5
            ):
                self.win8 = True
            self.local.usages = []
            self.local.usage_sizes = []
            self.local.usage_min = 0
            self.local.usage_max = 0
            self.local.usage_max_size = 0

    def dump(
        self: "ReportDescriptor", dump_file=sys.stdout, output_type="default"
    ) -> None:
        """
        Write this ReportDescriptor into the given file

        The "default" format prints each item as hexadecimal format with a
        double-slash comment, e.g. ::

           0x05, 0x01,                    // Usage Page (Generic Desktop)        0
           0x09, 0x02,                    // Usage (Mouse)                       2


        The "kernel" format prints each item in valid C format, for easy
        copy-paste into a kernel or C source file: ::

               0x05, 0x01,         /* Usage Page (Generic Desktop)         */
               0x09, 0x02,         /* Usage (Mouse)                        */

        :param File dump_file: the file to write to
        :param str output_type: the output format, one of "default" or "kernel"
        """
        assert output_type in ["default", "kernel", "human"]

        indent = 0
        for rdesc_item in self.rdesc_items:
            if output_type == "default":
                indent = rdesc_item.dump_rdesc_array(indent, dump_file)
            elif output_type == "kernel":
                indent = rdesc_item.dump_rdesc_kernel(indent, dump_file)
            elif output_type == "human":
                indent = rdesc_item.dump_rdesc_human(indent, dump_file)

    @property
    def size(self: "ReportDescriptor") -> int:
        """
        Returns the size of the report descriptor in bytes.
        """
        return sum([item.size for item in self.rdesc_items])

    @property
    def bytes(self: "ReportDescriptor") -> List[U8]:
        """
        This report descriptor as a list of 8-bit integers.
        """
        data = []
        for item in self.rdesc_items:
            data.extend(item.bytes)
        return data

    @classmethod
    def from_bytes(
        cls: _Type["ReportDescriptor"], rdesc: Union[Bytes, List[U8]]
    ) -> "ReportDescriptor":
        """
        Parse the given list of 8-bit integers.

        :param list rdesc: a list of bytes that are this report descriptor
        """
        items = _HidRDescItem.from_bytes(rdesc)

        return ReportDescriptor(items)

    @classmethod
    def from_string(cls: _Type["ReportDescriptor"], rdesc: str) -> "ReportDescriptor":
        """
        Parse a string in the format of series of hex numbers::

           12 34 ab cd ...

        and the first number in that series is the count of bytes, excluding
        that first number. This is the format returned by your
        ``/dev/hidraw`` event node, so just pass it along.


        :param list rdesc: a string that represents the list of bytes
        """

        irdesc = [int(r, 16) for r in rdesc.split()[1:]]
        items = _HidRDescItem.from_bytes(irdesc)

        return ReportDescriptor(items)

    @classmethod
    def from_human_descr(
        cls: _Type["ReportDescriptor"], rdesc_str: str
    ) -> "ReportDescriptor":
        """
        Parse the given human-readable report descriptor, e.g. ::

            Usage Page (Digitizers)
            Usage (Finger)
            Collection (Logical)
             Report Size (1)
             Report Count (1)
             Logical Minimum (0)
             Logical Maximum (1)
             Usage (Tip Switch)
             Input (Data,Var,Abs)
             Report Size (7)
             Logical Maximum (127)
             Input (Cnst,Var,Abs)
             Report Size (8)
             Logical Maximum (255)
             Usage (Contact Id)

        """
        usage_page = 0
        items = []
        for line in rdesc_str.splitlines():
            line = line.strip()
            if not line:
                continue
            item = _HidRDescItem.from_human_descr(line, usage_page)
            usage_page = item.usage_page >> 16
            items.append(item)

        return ReportDescriptor(items)


def get_descriptor(fd: int):
    from hhd.controller.lib.ioctl import HIDIOCGRDESC, HIDIOCGRDESCSIZE
    from fcntl import ioctl
    import ctypes

    size = ctypes.c_int32()
    ioctl(fd, HIDIOCGRDESCSIZE, size)
    c_mask = ctypes.create_string_buffer(4096+4)
    c_mask[:4] = size.value.to_bytes(4, byteorder=sys.byteorder)
    ioctl(fd, HIDIOCGRDESC, c_mask)
    rdesc = c_mask.raw
    return rdesc[4:size.value+4]

def print_descriptor(fd: int, format: str = 'default'):
    desc = get_descriptor(fd)
    return ReportDescriptor.from_bytes(desc).dump(output_type=format)



from .base import (
    Consumer,
    Event,
    KeyboardWrapper,
    Multiplexer,
    Producer,
    TouchpadCorrection,
    TouchpadCorrectionType,
    can_read,
    correct_touchpad,
    SpecialEvent,
    ControllerEmitter,
    DEBUG_MODE,
    RgbMode,
    RgbCapabilities,
)
from .const import Axis, Button, Configuration

__all__ = [
    "Axis",
    "Button",
    "Event",
    "Configuration",
    "Consumer",
    "Producer",
    "can_read",
    "TouchpadCorrectionType",
    "TouchpadCorrection",
    "correct_touchpad",
    "KeyboardWrapper",
    "Multiplexer",
    "SpecialEvent",
    "ControllerEmitter",
    "DEBUG_MODE",
    "RgbMode",
    "RgbCapabilities",
]



SDCONT_VENDOR = 0x28DE
SDCONT_PRODUCT = 0x1205
SDCONT_VERSION = 256
SDCONT_COUNTRY = 0
SDCONT_NAME = b"Emulated Steam Controller"

SDCONT_DESCRIPTOR = bytes(
    [
        0x06,
        0xFF,
        0xFF,  # // Usage Page (Vendor Usage Page 0xffff) 0
        0x09,
        0x01,  # // Usage (Vendor Usage 0x01)           3
        0xA1,
        0x01,  # // Collection (Application)            5
        0x09,
        0x02,  # //  Usage (Vendor Usage 0x02)          7
        0x09,
        0x03,  # //  Usage (Vendor Usage 0x03)          9
        0x15,
        0x00,  # //  Logical Minimum (0)                11
        0x26,
        0xFF,
        0x00,  # //  Logical Maximum (255)              13
        0x75,
        0x08,  # //  Report Size (8)                    16
        0x95,
        0x40,  # //  Report Count (64)                  18
        0x81,
        0x02,  # //  Input (Data,Var,Abs)               20
        0x09,
        0x06,  # //  Usage (Vendor Usage 0x06)          22
        0x09,
        0x07,  # //  Usage (Vendor Usage 0x07)          24
        0x15,
        0x00,  # //  Logical Minimum (0)                26
        0x26,
        0xFF,
        0x00,  # //  Logical Maximum (255)              28
        0x75,
        0x08,  # //  Report Size (8)                    31
        0x95,
        0x40,  # //  Report Count (64)                  33
        0xB1,
        0x02,  # //  Feature (Data,Var,Abs)             35
        0xC0,  # // End Collection                      37
    ]
)



import itertools
import json
import logging
import os
from copy import deepcopy
from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn
from threading import Condition, Thread
from typing import Any, Mapping, Sequence, cast
from urllib.parse import parse_qs, urlparse

from hhd.plugins import (
    Config,
    Context,
    Emitter,
    HHDLocale,
    HHDSettings,
    get_relative_fn,
    load_relative_yaml,
)

from .i18n import get_user_lang, translate, translate_ver

logger = logging.getLogger(__name__)


def sanitize_name(n: str):
    import re

    return re.sub(r"[^ a-zA-Z0-9]+", "", n)


def sanitize_fn(n: str):
    import re

    return re.sub(r"[^ a-zA-Z0-9\._/]+", "", n)


STANDARD_HEADERS = {
    "Access-Control-Allow-Origin": "*",
    "Access-Control-Allow-Credentials": "true",
    "Access-Control-Allow-Methods": "POST, GET, OPTIONS, DELETE",
    "Access-Control-Allow-Headers": "*",
    # "Access-Control-Expose-Headers": "*, Version",
    "Access-Control-Max-Age": "86400",
    "WWW-Authenticate": "Bearer",
}

ERROR_HEADERS = {**STANDARD_HEADERS, "Content-type": "text/plain"}
AUTH_HEADERS = ERROR_HEADERS
OK_HEADERS = {**STANDARD_HEADERS, "Content-type": "application/json"}

# https://en.wikipedia.org/wiki/List_of_Unicode_characters#Control_codes
_control_char_table = str.maketrans(
    {c: rf"\x{c:02x}" for c in itertools.chain(range(0x20), range(0x7F, 0xA0))}
)
_control_char_table[ord("\\")] = r"\\"

SECTIONS = load_relative_yaml("../sections.yml")["sections"]


def parse_path(path: str) -> tuple[list, dict[str, list[str]]]:
    try:
        url = urlparse(path)
        if url.path:
            segments = url.path[1:].split("/")
        else:
            segments = []

        params = {
            k: v for k, v in parse_qs(url.query, keep_blank_values=True).items() if v
        }
        return segments, params
    except Exception:
        return [], {}


class RestHandler(BaseHTTPRequestHandler):
    settings: HHDSettings
    cond: Condition
    conf: Config
    info: Config
    profiles: Mapping[str, Config]
    emit: Emitter
    locales: Sequence[HHDLocale]
    ctx: Context
    user_lang: str | None
    token: str | None

    def set_response(self, code: int, headers: dict[str, str] = {}):
        # Allow skipping CORS by responding with specific origin
        if og := self.headers.get("Origin", None):
            headers = {**headers, "Access-Control-Allow-Origin": og}
        self.send_response(code)
        for title, head in headers.items():
            self.send_header(title, head)
        self.end_headers()

    def do_OPTIONS(self):
        self.set_response(
            204,
            STANDARD_HEADERS,
        )

    def is_authenticated(self):
        if not self.token:
            return True

        auth = self.headers["Authorization"]
        if not auth:
            return False

        if not isinstance(auth, str):
            return False

        if not auth.lower().startswith("bearer "):
            return False

        return auth.lower()[len("Bearer ") :] == self.token

    def send_authenticate(self):
        if self.is_authenticated():
            return True

        self.set_response(401, {"Content-type": "text/plain"})
        self.wfile.write(
            f"Handheld Daemon Error: Authentication is on and you did not supply the proper bearer token.".encode()
        )

        return False

    def send_json(self, data: Any):
        self.set_response_ok()
        self.wfile.write(json.dumps(data).encode())

    def set_response_ok(self, extra_headers={}):
        self.set_response(200, {**OK_HEADERS, **extra_headers})

    def send_not_found(self, error: str):
        self.set_response(404, ERROR_HEADERS)
        self.wfile.write(b"Handheld Daemon Error (404, invalid endpoint):\n")
        self.wfile.write(error.encode())

    def send_error_str(self, error: str):
        self.set_response(400, ERROR_HEADERS)
        self.wfile.write(b"Handheld Daemon Error:\n")
        self.wfile.write(error.encode())

    def send_error(self, *args, **kwargs):
        if len(args) == 1:
            return self.send_error_str(args[0])
        else:
            for title, head in STANDARD_HEADERS.items():
                self.send_header(title, head)
            return super().send_error(*args, **kwargs)

    def send_file(self, fn: str):
        if not "." in fn:
            return self.send_error(f"Invalid file: {fn}")
        match fn[fn.rindex(".") :]:
            case ".css":
                ctype = "text/css"
            case ".js":
                ctype = "application/javascript"
            case ".html" | ".htm" | ".php":
                ctype = "text/html"
            case other:
                return self.send_error(f"File type '{other} of '{fn}' not supported.")
        self.set_response(200, {**STANDARD_HEADERS, "Content-type": ctype})
        with open(get_relative_fn(fn), "rb") as f:
            self.wfile.write(f.read())

    def handle_profile(
        self, segments: list[str], params: dict[str, list[str]], content: Any | None
    ):
        if not segments:
            return self.send_not_found(
                f"No endpoint provided for '/profile/...', (e.g., list, get, set, apply)"
            )

        with self.cond:
            match segments[0]:
                case "list":
                    self.send_json(list(self.profiles))
                case "get":
                    if "profile" not in params:
                        return self.send_error(f"Profile not specified")
                    profile = sanitize_name(params["profile"][0])
                    if profile not in self.profiles:
                        return self.send_error(f"Profile '{profile}' not found.")
                    self.send_json(self.profiles[profile].conf)
                case "set":
                    if "profile" not in params:
                        return self.send_error(f"Profile not specified")
                    if not content or not isinstance(content, Mapping):
                        return self.send_error(f"Data for the profile not sent.")

                    profile = sanitize_name(params["profile"][0])
                    self.emit(
                        {"type": "profile", "name": profile, "config": Config(content)}
                    )
                    # Wait for the profile to be processed
                    self.cond.wait()

                    # Return the profile
                    if profile in self.profiles:
                        self.send_json(self.profiles[profile].conf)
                    else:
                        self.send_error(f"Applied profile not found (race condition?).")
                case "del":
                    if "profile" not in params:
                        return self.send_error(f"Profile not specified")

                    profile = sanitize_name(params["profile"][0])
                    if profile not in self.profiles:
                        return self.send_error(f"Profile '{profile}' not found.")
                    self.emit({"type": "profile", "name": profile, "config": None})
                    # Wait for the profile to be processed
                    self.cond.wait()

                    if profile in self.profiles:
                        self.send_error(f"Applied profile not found (race condition?).")
                    else:
                        self.set_response_ok()
                case "apply":
                    if "profile" not in params:
                        return self.send_error(f"Profile not specified")

                    profiles = [sanitize_name(p) for p in params["profile"]]
                    for p in profiles:
                        if p not in self.profiles:
                            return self.send_error(f"Profile '{p}' not found.")

                    self.emit([{"type": "apply", "name": p} for p in profiles])
                    # Wait for the profile to be processed
                    self.cond.wait()
                    # Return the profile
                    self.send_json(self.conf.conf)
                case other:
                    self.send_not_found(f"Command 'profile/{other}' not supported.")

    def handle_image(
        self, segments: list[str], params: dict[str, list[str]], content: Any | None
    ):
        if len(segments) < 2:
            return self.send_not_found(
                "Image data not provided. Syntax: /api/v1/image/{game}/{type}"
            )

        game = segments[0]
        image_type = segments[1]

        img = self.emit.get_image(game, image_type)
        if img is None:
            return self.send_error(f"Image '{game}/{image_type}' not found.")

        if img.endswith(".jpg") or img.endswith(".jpeg"):
            ctype = "image/jpeg"
        elif img.endswith(".png"):
            ctype = "image/png"
        else:
            return self.send_error(f"Image type '{img}' not supported.")

        if not os.path.exists(img):
            return self.send_error(f"Image '{img}' not found.")

        self.set_response_ok(extra_headers={"Content-type": ctype})
        with open(img, "rb") as f:
            self.wfile.write(f.read())

    def v1_endpoint(self, content: Any | None):
        segments, params = parse_path(self.path)
        langs = params.get("lang", params.get("locale", None))
        lang = langs[0] if langs else None

        if not segments:
            return self.send_not_found(f"Empty path.")

        if segments[0] != "api":
            return self.send_not_found(
                f"Only the API endpoint ('/api/v1') is supported for now."
            )

        if len(segments) < 2 or segments[1] != "v1":
            return self.send_not_found(
                f"Only v1 endpoint is supported by this version of hhd ('/api/v1')."
            )

        if len(segments) == 2:
            return self.send_not_found(f"No command provided")

        command = segments[2].lower()
        match command:
            case "image" | "images":
                self.handle_image(segments[3:], params, content)
            case "profile":
                self.handle_profile(segments[3:], params, content)
            case "settings":
                v = translate_ver(self.conf, lang=lang, user_lang=self.user_lang)
                self.set_response_ok({"Version": v})
                with self.cond:
                    s = dict(deepcopy(self.settings))
                    try:
                        s["hhd"]["version"] = {  # type: ignore
                            "type": "version",
                            "tags": ["non-essential", "advanced", "expert", "hide"],
                            "value": v,
                        }
                    except Exception as e:
                        logger.error(f"Error while writing version hash to response.")
                    s = translate(
                        s, self.conf, self.locales, lang=lang, user_lang=self.user_lang
                    )
                    self.wfile.write(json.dumps(s).encode())
            case "state":
                self.set_response_ok()
                with self.cond:
                    if content:
                        if not isinstance(content, Mapping):
                            return self.send_error(
                                f"State content should be a dictionary."
                            )
                        self.emit({"type": "state", "config": Config(content)})
                        self.cond.wait()
                    elif "poll" in params:
                        # Hang for the next update if the UI requests it.
                        self.cond.wait()
                    out = {**cast(dict, self.conf.conf), "info": self.info.conf}
                    out["version"] = translate_ver(
                        self.conf, lang=lang, user_lang=self.user_lang
                    )
                    out = translate(
                        out,
                        self.conf,
                        self.locales,
                        lang=lang,
                        user_lang=self.user_lang,
                    )
                    self.wfile.write(json.dumps(out).encode())
            case "version":
                self.send_json({"version": 5})
            case "sections":
                self.send_json(
                    translate(
                        SECTIONS,
                        self.conf,
                        self.locales,
                        lang=lang,
                        user_lang=self.user_lang,
                    )
                )
            case other:
                self.send_not_found(f"Command '{other}' not supported.")

    def do_GET(self):
        # Danger zone unauthenticated
        # Be very careful
        try:
            path = sanitize_fn(urlparse(self.path).path)
            if path.startswith("/"):
                path = path[1:]
            match path.split("/"):
                case ["" | "index.html" | "index.php"]:
                    return self.send_file("./index.html")
                case ["static", *other]:
                    return self.send_file(os.path.join("static", *other))
                case ["api", *other]:
                    if not self.send_authenticate():
                        return
                    self.v1_endpoint(None)
                case other:
                    return self.send_not_found(f"File not found:\n{path}")
        except Exception as e:
            logger.debug(f"Encountered error while serving unauthenticated request.")
            try:
                return self.send_error(f"Encountered error while serving request:\n{e}")
            except Exception:
                # Generated due to polling website going in the background
                pass

    def do_POST(self):
        if not self.send_authenticate():
            return

        content_length = int(self.headers["Content-Length"])
        content = self.rfile.read(content_length)
        try:
            content_json = json.loads(content)
        except Exception as e:
            return self.send_error(
                f"Parsing the POST content as json failed with the following error:\n{e}"
            )
        self.v1_endpoint(content_json)

    def log_message(self, format: str, *args: Any) -> None:
        message = format % args
        logger.error(
            f"Received invalid request from '{self.address_string()}':\n{message.translate(_control_char_table)}"
        )

    def log_request(self, code="-", size="-"):
        pass

    def __getattr__(self, val: str):
        if not val.startswith("do_"):
            raise AttributeError()

        logger.warning(
            f"Received request type '{val[3:].translate(_control_char_table)}' from '{self.address_string()}'. Handling as GET."
        )
        return self.do_GET


class ThreadingSimpleServer(ThreadingMixIn, HTTPServer):
    pass


class HHDHTTPServer:
    def __init__(
        self,
        localhost: bool,
        port: int,
        token: str | None,
    ) -> None:
        self.localhost = localhost
        self.port = port

        # Have to subclass to create closure
        class NewRestHandler(RestHandler):
            pass

        cond = Condition()
        NewRestHandler.cond = cond
        NewRestHandler.token = token
        self.cond = cond
        self.handler = NewRestHandler
        self.https = None
        self.t = None

    def update(
        self,
        settings: HHDSettings,
        conf: Config,
        info: Config,
        profiles: Mapping[str, Config],
        emit: Emitter,
        locales: Sequence[HHDLocale],
        ctx: Context,
    ):
        with self.cond:
            self.handler.settings = settings
            self.handler.conf = conf
            self.handler.info = info
            self.handler.profiles = profiles
            self.handler.emit = emit
            self.handler.locales = locales
            self.handler.ctx = ctx
            if not hasattr(self.handler, "user_lang"):
                # Only load user lang once to avoid weirdness
                self.handler.user_lang = get_user_lang(ctx)
            self.cond.notify_all()

    def open(self):
        self.https = ThreadingSimpleServer(
            ("127.0.0.1" if self.localhost else "", self.port), self.handler
        )
        self.t = Thread(target=self.https.serve_forever)
        self.t.start()

    def close(self):
        if self.https and self.t:
            with self.cond:
                self.cond.notify_all()
            self.https.shutdown()
            self.t.join()
            self.https = None
            self.t = None



import datetime
import logging
import os
import pathlib
from logging.handlers import RotatingFileHandler
from typing import Sequence, Any

from rich.logging import RichHandler
from threading import local, Lock, get_ident, enumerate
from .utils import Context, expanduser, fix_perms

logger = logging.getLogger(__name__)

RASTER = """\
   _______   _______    ______  
  /    /  ╲╲/    /  ╲╲_/      ╲╲
 /        //        //        //
/         /         /         / 
╲___/____/╲___/____/╲________/  \n"""


class NewLineFormatter(logging.Formatter):
    """Aligns newlines during multiline prints."""

    def format(self, record):
        msg = super().format(record)
        if (idx := msg.index("|||")) != -1:
            preamble = msg[:idx]
            msg = msg.replace("|||", "").replace("\n", "\n" + (" " * len(preamble)))
        return msg


_lock = Lock()
_main = "main"
_plugins = {}


def set_log_plugin(plugin: str = "main"):
    global _main
    with _lock:
        _plugins[get_ident()] = plugin
        _main = plugin


def get_log_plugin():
    with _lock:
        return _plugins.get(get_ident(), _main)


def update_log_plugins():
    for t in enumerate():
        if t.ident and t.ident not in _plugins:
            _plugins[t.ident] = _main


class PluginLogRender:
    def __init__(
        self,
        time_format="[%x %X]",
        omit_repeated_times: bool = True,
        level_width=8,
        plugin_width=5,
        print_time=True,
        print_path=True,
    ) -> None:
        from rich.style import Style

        self.time_format = time_format
        self.omit_repeated_times = omit_repeated_times
        self.level_width = level_width
        self._last_time = None
        self.plugin_width = plugin_width
        self.print_time = print_time
        self.print_path = print_path

    def __call__(
        self,
        console,
        renderables,
        log_time=None,
        time_format=None,
        level: Any = "",
        plugin: str | None = None,
        path: str | None = None,
        line_no: int | None = None,
        link_path: str | None = None,
    ):
        from rich.containers import Renderables
        from rich.table import Table
        from rich.text import Text

        output = Table.grid(padding=(0, 1))
        output.expand = True
        if self.print_time:
            output.add_column(style="log.time")
        match plugin:
            case "main":
                color = "magenta"
            case "ukwn":
                color = "red"
            case _:
                color = "cyan"
        output.add_column(style=color, width=self.plugin_width)
        output.add_column(style="log.level", width=self.level_width)

        output.add_column(ratio=1, style="log.message", overflow="fold")
        if self.print_path:
            output.add_column(style="log.path")

        row = []
        if self.print_time:
            log_time = log_time or console.get_datetime()
            time_format = time_format or self.time_format
            if callable(time_format):
                log_time_display = time_format(log_time)
            else:
                log_time_display = Text(log_time.strftime(time_format))
            if log_time_display == self._last_time and self.omit_repeated_times:
                row.append(Text(" " * len(log_time_display)))
            else:
                row.append(log_time_display)
                self._last_time = log_time_display
        row.append(plugin.upper() if plugin else "")
        row.append(level)

        # Find plugin
        row.append(Renderables(renderables))
        if path and self.print_path:
            path_text = Text()
            path_text.append(
                path, style=f"link file://{link_path}" if link_path else ""
            )
            if line_no:
                path_text.append(":")
                path_text.append(
                    f"{line_no}",
                    style=f"link file://{link_path}#{line_no}" if link_path else "",
                )
            row.append(path_text)

        output.add_row(*row)
        return output


class PluginRichHandler(RichHandler):
    def __init__(self, renderer: PluginLogRender) -> None:
        self.renderer = renderer
        super().__init__()

    def render(
        self,
        *,
        record,
        traceback,
        message_renderable,
    ):
        path = pathlib.Path(record.pathname).name
        level = self.get_level_text(record)
        time_format = None if self.formatter is None else self.formatter.datefmt
        log_time = datetime.datetime.fromtimestamp(record.created)

        log_renderable = self.renderer(
            self.console,
            [message_renderable] if not traceback else [message_renderable, traceback],
            log_time=log_time,
            time_format=time_format,
            level=level,
            plugin=get_log_plugin(),
            path=path,
            line_no=record.lineno,
            link_path=record.pathname if self.enable_link_path else None,
        )
        return log_renderable


class UserRotatingFileHandler(RotatingFileHandler):
    def __init__(
        self,
        filename: str,
        mode: str = "a",
        maxBytes: int = 0,
        backupCount: int = 0,
        encoding: str | None = None,
        delay: bool = False,
        errors: str | None = None,
        ctx: Context | None = None,
    ) -> None:
        self.ctx = ctx
        super().__init__(filename, mode, maxBytes, backupCount, encoding, delay, errors)

    def _open(self):
        d = super()._open()
        if self.ctx:
            os.chown(self.baseFilename, self.ctx.euid, self.ctx.egid)
        return d


def setup_logger(
    log_dir: str | None = None, init: bool = True, ctx: Context | None = None
):
    from rich import get_console
    from rich.traceback import install

    if log_dir:
        log_dir = expanduser(log_dir, ctx)

    # Do not print time when running as a systemd service
    is_systemd = bool(os.environ.get("JOURNAL_STREAM", None))

    install()
    handlers = []
    handlers.append(
        PluginRichHandler(
            PluginLogRender(print_time=not is_systemd, print_path=not is_systemd)
        )
    )
    if log_dir:
        os.makedirs(log_dir, exist_ok=True)
        if ctx:
            fix_perms(log_dir, ctx)
        handler = UserRotatingFileHandler(
            os.path.join(log_dir, "hhd.log"),
            maxBytes=10_000_000,
            backupCount=10,
            ctx=ctx,
        )
        handler.setFormatter(
            NewLineFormatter("%(asctime)s %(module)-15s %(levelname)-8s|||%(message)s")
        )
        handler.doRollover()
        handlers.append(handler)

    FORMAT = "%(message)s"
    logging.basicConfig(
        level=logging.INFO,
        datefmt="[%H:%M]",
        format=FORMAT,
        handlers=handlers,
    )
    if init:
        get_console().print(RASTER, justify="full", markup=False, highlight=False)
        logger.info(f"Handheld Daemon starting...")



from threading import Event, Thread
from typing import Any, Sequence

from hhd.plugins import (
    Config,
    Context,
    Emitter,
    HHDPlugin,
    load_relative_yaml,
    get_outputs_config,
    get_touchpad_config,
    get_gyro_state,
    get_gyro_config,
)
from hhd.plugins.settings import HHDSettings
from hhd.controller.physical.imu import BMI_MAPPINGS


class LegionControllersPlugin(HHDPlugin):
    name = "legion_go_controllers"
    priority = 18
    log = "llgo"

    def __init__(self) -> None:
        self.t = None
        self.should_exit = None
        self.updated = Event()
        self.started = False
        self.t = None
        self.prev = None

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        self.emit = emit
        self.context = context
        self.prev = None

    def settings(self) -> HHDSettings:
        base = {"controllers": {"legion_go": load_relative_yaml("controllers.yml")}}
        base["controllers"]["legion_go"]["children"]["xinput"].update(
            get_outputs_config(extra_buttons='quad')
        )
        base["controllers"]["legion_go"]["children"]["touchpad"] = get_touchpad_config()
        return base

    def update(self, conf: Config):
        new_conf = conf["controllers.legion_go"]
        reset = conf["controllers.legion_go.factory_reset"].to(bool)
        conf["controllers.legion_go.factory_reset"] = False

        # Migrate old setting
        val = conf.get("controllers.legion_go.swap_legion", None)
        if val and val != "disabled":
            conf["controllers.legion_go.swap_legion"] = None
            conf["controllers.legion_go.swap_legion_v2"] = True

        if new_conf == self.prev:
            return
        if self.prev is None:
            self.prev = new_conf
        else:
            self.prev.update(new_conf.conf)

        if reset:
            self.started = False
        else:
            self.updated.set()
        self.start(self.prev, reset)

    def start(self, conf, reset=False):
        from .base import plugin_run

        if self.started:
            return
        self.started = True

        self.close()
        self.should_exit = Event()
        self.t = Thread(
            target=plugin_run,
            args=(
                conf,
                self.emit,
                self.context,
                self.should_exit,
                self.updated,
                {"reset": reset},
            ),
        )
        self.t.start()

    def close(self):
        if not self.should_exit or not self.t:
            return
        self.should_exit.set()
        self.t.join()
        self.should_exit = None
        self.t = None


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    # Match just product number, should be enough for now
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        if not f.read().strip() == "83E1":
            return []

    return [LegionControllersPlugin()]



from typing import Any, Sequence, TYPE_CHECKING
import os
from hhd.plugins import (
    HHDPlugin,
    Context,
)
from time import sleep
from hhd.plugins import HHDSettings, load_relative_yaml
import logging

from hhd.plugins.conf import Config

logger = logging.getLogger(__name__)
BACKLIGHT_DIR = "/sys/class/backlight/"


def write_sysfs(dir: str, fn: str, val: Any):
    with open(os.path.join(dir, fn), "w") as f:
        f.write(str(val))


def read_sysfs(dir: str, fn: str, default: str | None = None):
    try:
        with open(os.path.join(dir, fn), "r") as f:
            return f.read().strip()
    except Exception as e:
        if default is not None:
            return default
        raise e


class DisplayPlugin(HHDPlugin):
    def __init__(self) -> None:
        self.name = f"displayd"
        self.priority = 4
        self.log = "disp"

        self.display = None
        self.max_brightness = 255

    def settings(self) -> HHDSettings:
        if self.display:
            return {"system": {"display": load_relative_yaml("settings.yml")}}
        else:
            return {}

    def open(
        self,
        emit,
        context: Context,
    ):
        self.display = None
        self.prev = None
        for d in os.listdir(BACKLIGHT_DIR):
            ddir = os.path.join(BACKLIGHT_DIR, d)
            try:
                read_sysfs(ddir, "brightness")
                max_bright = int(read_sysfs(ddir, "max_brightness"))
                self.display = ddir
                self.max_brightness = max_bright
            except Exception:
                pass

        if self.display is None:
            logger.warning(f"Display with variable brightness not found. Exitting.")

    def update(self, conf: Config):
        if not self.display:
            return

        curr = None
        try:
            requested = conf["system.display.brightness"].to(int)

            curr = int(
                int(read_sysfs(self.display, "brightness", None))
                * 100
                / self.max_brightness
            )

            # Set brightness
            if requested is not None and requested != self.prev:
                changed = False
                # If the change is too low the display might not make the
                # change, so while loop and increase requested values
                logger.info(f"Setting brightness to {requested}")
                while not changed and (requested >= 0 and requested <= 100):
                    write_sysfs(
                        self.display,
                        "brightness",
                        int(self.max_brightness * requested / 100),
                    )

                    # Get brightness
                    new_curr = int(
                        int(read_sysfs(self.display, "brightness", None))
                        * 100
                        / self.max_brightness
                    )
                    changed = new_curr != curr
                    curr = new_curr

                    # In case the brightness did not change
                    # increase request
                    requested_old = requested
                    if curr > requested:
                        requested -= 1
                    else:
                        requested += 1

                    if not changed:
                        logger.warning(
                            f"Could not set brightness to {requested_old}. Trying {requested}."
                        )

            conf["general.display.brightness"] = curr
            self.prev = curr
        except Exception as e:
            logger.error(f"Error while processing display settings:\n{type(e)}: {e}")
            # Set conf to avoid repeated updates
            conf["general.display.brightness"] = curr

    def close(self):
        pass


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    return [DisplayPlugin()]



from typing import NamedTuple


class TouchScreenQuirk(NamedTuple):
    portrait: bool
    flip_x: bool  # Left <-> Right
    flip_y: bool  # Top <-> Bottom


TQ = TouchScreenQuirk


class TouchScreenMatch(NamedTuple):
    dmi: str | None = None
    vid: int | None = None
    pid: int | None = None
    name: str | None = None

DEFAULT_LANDSCAPE = TQ(False, True, False)

TM = TouchScreenMatch

TOUCH_SCREEN_QUIRKS = {
    # Lenovo
    TM("83E1", name="Legion GO"): TQ(True, False, False),
    # MinisForum
    TM("V3", name="MinisForum V3"): DEFAULT_LANDSCAPE,
    # Steam deck
    TM("Jupiter", name="Steam Deck LCD"): TQ(True, True, True),
    TM("Galileo", name="Steam Deck OLED"): TQ(True, True, True),
    # GPD
    TM("G1618-04", name="GPD Win 4"): DEFAULT_LANDSCAPE,  # 2023: 0x0416:0x038F
    TM("G1619-04", name="GPD Win Max 2 (04)"): DEFAULT_LANDSCAPE,  # 2023: 27C6:0113
    TM("G1619-05", name="GPD Win Max 2 (05)"): DEFAULT_LANDSCAPE,
    # Asus
    TM("RC71L", name="ROG Ally"): DEFAULT_LANDSCAPE,
    TM("RC72LA", name="ROG Ally X"): DEFAULT_LANDSCAPE,
}


def get_touchscreen_quirk(vid=None, pid=None):
    try:
        with open("/sys/class/dmi/id/product_name") as f:
            dmi = f.read().strip()
    except Exception:
        return None, None

    for match, quirk in TOUCH_SCREEN_QUIRKS.items():
        if match.dmi and match.dmi not in dmi:
            continue
        if match.vid and match.vid != vid:
            continue
        if match.pid and match.pid != pid:
            continue

        return (
            quirk,
            match.name,
        )

    return None, None


def get_system_info():
    try:
        with open("/sys/class/dmi/id/product_name") as f:
            dmi = f.read().strip()
    except Exception:
        dmi = ""

    try:
        with open("/sys/class/dmi/id/sys_vendor") as f:
            vendor = f.read().strip()
    except Exception:
        vendor = ""

    return vendor, dmi



from typing import cast, Literal

from hhd.controller import Axis

from .conf import Config
from .utils import load_relative_yaml


def get_product():
    try:
        with open("/sys/devices/virtual/dmi/id/product_name", "r") as f:
            return f.read().strip()
    except Exception:
        return "Uknown"


def get_vendor():
    try:
        with open("/sys/devices/virtual/dmi/id/board_vendor", "r") as f:
            return f.read().strip()
    except Exception:
        return "Uknown"


def get_touchpad_config():
    return load_relative_yaml("touchpad.yml")


def get_gyro_config(
    mapping: dict[str, tuple[Axis, str | None, float, float | None]] | None
):
    g = load_relative_yaml("gyro.yml")
    g["modes"]["remapped"]["children"]["manufacturer"]["default"] = f'"{get_vendor()}"'
    g["modes"]["remapped"]["children"]["product"]["default"] = f'"{get_product()}"'
    if mapping:
        for key, (ax, _, scale, _) in mapping.items():
            match key:
                case "anglvel_x":
                    setting = "x"
                case "anglvel_y":
                    setting = "y"
                case "anglvel_z":
                    setting = "z"
                case _:
                    setting = None
            match ax:
                case "gyro_x":
                    default = "x"
                case "gyro_y":
                    default = "y"
                case "gyro_z":
                    default = "z"
                case _:
                    default = None
            invert = scale < 0

            if setting and default:
                g["modes"]["remapped"]["children"][f"{setting}_axis"][
                    "default"
                ] = default
                g["modes"]["remapped"]["children"][f"{setting}_invert"][
                    "default"
                ] = invert
    return g


def get_gyro_state(
    conf: Config,
    default: dict[str, tuple[Axis, str | None, float, float | None]],
) -> dict[str, tuple[Axis, str | None, float, float | None]]:
    if conf["mode"].to(str) == "default":
        return default

    rem = conf.get("remapped", {})
    return {
        "timestamp": ("imu_ts", None, 1, None),
        "accel_x": (
            cast(Axis, f"accel_{rem.get('x_axis', 'x')}"),
            "accel",
            -1 if rem.get("x_invert", False) else 1,
            3,
        ),
        "accel_y": (
            cast(Axis, f"accel_{rem.get('y_axis', 'y')}"),
            "accel",
            -1 if rem.get("y_invert", False) else 1,
            3,
        ),
        "accel_z": (
            cast(Axis, f"accel_{rem.get('z_axis', 'z')}"),
            "accel",
            -1 if rem.get("z_invert", False) else 1,
            3,
        ),
        "anglvel_x": (
            cast(Axis, f"gyro_{rem.get('x_axis', 'x')}"),
            "anglvel",
            -1 if rem.get("x_invert", False) else 1,
            None,
        ),
        "anglvel_y": (
            cast(Axis, f"gyro_{rem.get('y_axis', 'y')}"),
            "anglvel",
            -1 if rem.get("y_invert", False) else 1,
            None,
        ),
        "anglvel_z": (
            cast(Axis, f"gyro_{rem.get('z_axis', 'z')}"),
            "anglvel",
            -1 if rem.get("z_invert", False) else 1,
            None,
        ),
    }


AxChoice = Literal["x", "y", "z"]


def gen_gyro_state(
    x: AxChoice, inv_x: bool, y: AxChoice, inv_y: bool, z: AxChoice, inv_z: bool
):
    return {
        "timestamp": ("imu_ts", None, 1, None),
        "accel_x": (
            cast(Axis, f"accel_{x}"),
            "accel",
            -1 if inv_x else 1,
            None,
        ),
        "accel_y": (
            cast(Axis, f"accel_{y}"),
            "accel",
            -1 if inv_y else 1,
            None,
        ),
        "accel_z": (
            cast(Axis, f"accel_{z}"),
            "accel",
            -1 if inv_z else 1,
            None,
        ),
        "anglvel_x": (
            cast(Axis, f"gyro_{x}"),
            "anglvel",
            -1 if inv_x else 1,
            None,
        ),
        "anglvel_y": (
            cast(Axis, f"gyro_{y}"),
            "anglvel",
            -1 if inv_y else 1,
            None,
        ),
        "anglvel_z": (
            cast(Axis, f"gyro_{z}"),
            "anglvel",
            -1 if inv_z else 1,
            None,
        ),
    }



"""
SPDX-License-Identifier: MIT
From: https://github.com/ValvePython/steam/
With fix from: https://github.com/solsticegamestudios/steam/
"""

"""
Appache file parsing examples:

.. code:: python

    >>> from steam.utils.appcache import parse_appinfo, parse_packageinfo

    >>> header, apps = parse_appinfo(open('/d/Steam/appcache/appinfo.vdf', 'rb'))
    >>> header
    {'magic': b")DV\\x07", 'universe': 1}
    >>> next(apps)
    {'appid': 5,
     'size': 79,
     'info_state': 1,
     'last_updated': 1484735377,
     'access_token': 0,
     'sha1': b'\\x87\\xfaCg\\x85\\x80\\r\\xb4\\x90Im\\xdc}\\xb4\\x81\\xeeQ\\x8b\\x825',
     'change_number': 4603827,
     'data_sha1': b'\\x87\\xfaCg\\x85\\x80\\r\\xb4\\x90Im\\xdc}\\xb4\\x81\\xeeQ\\x8b\\x825',
     'data': {'appinfo': {'appid': 5, 'public_only': 1}}}

    >>> header, pkgs = parse_packageinfo(open('/d/Steam/appcache/packageinfo.vdf', 'rb'))
    >>> header
    {'magic': b"'UV\\x06", 'universe': 1}

    >>> next(pkgs)
    {'packageid': 7,
     'sha1': b's\\x8b\\xf7n\\t\\xe5 k#\\xb6-\\x82\\xd2 \\x14k@\\xfeDQ',
     'change_number': 7469765,
     'data': {'7': {'packageid': 7,
       'billingtype': 1,
       'licensetype': 1,
       'status': 0,
       'extended': {'requirespreapproval': 'WithRedFlag'},
       'appids': {'0': 10, '1': 80, '2': 100, '3': 254430},
       'depotids': {'0': 0, '1': 95, '2': 101, '3': 102, '4': 103, '5': 254431},
       'appitems': {}}}}

"""

import struct
from .vdf import binary_load

uint32 = struct.Struct('<I')
uint64 = struct.Struct('<Q')
int64 = struct.Struct('<q')

def parse_appinfo(fp, mapper=None):
    """Parse appinfo.vdf from the Steam appcache folder

    :param fp: file-like object
    :param mapper: Python object class to return
    :raises: SyntaxError
    :rtype: (:class:`Generator` returning :class:`dict` by default or mapper class if set)
    :return: (header, apps iterator)
    """
# format:
#   uint32   - MAGIC: "'DV\x07" or "(DV\x07" or b")DV\x07"
#   uint32   - UNIVERSE: 1
#   int64    - OFFSET TO KEY TABLE (added in ")DV\x07")
#   ---- repeated app sections ----
#   uint32   - AppID
#   uint32   - size
#   uint32   - infoState
#   uint32   - lastUpdated
#   uint64   - accessToken
#   20bytes  - SHA1
#   uint32   - changeNumber
#   20bytes  - binary_vdf SHA1 (added in "(DV\x07")
#   variable - binary_vdf
#   ---- end of section ---------
#   uint32   - EOF: 0
#
#   ---- key table ----
#   uint32   - Count of keys
#   char[]   - Null-terminated strings corresponding to field names

    magic = fp.read(4)
    if magic not in (b"'DV\x07", b"(DV\x07", b")DV\x07"):
        raise SyntaxError("Invalid magic, got %s" % repr(magic))

    universe = uint32.unpack(fp.read(4))[0]

    key_table = None

    appinfo_version = magic[0]
    if appinfo_version >= 41:  # b')'
        # appinfo.vdf V29 and newer store list of keys in separate table at the
        # end of the file to reduce size. Retrieve it and pass it to the VDF
        # parser later.
        key_table = []

        key_table_offset = struct.unpack('q', fp.read(8))[0]
        offset = fp.tell()
        fp.seek(key_table_offset)
        key_count = uint32.unpack(fp.read(4))[0]

        # Read all null-terminated strings into a list
        for _ in range(0, key_count):
            field_name = bytearray()
            while True:
                field_name += fp.read(1)

                if field_name[-1] == 0:
                    field_name = field_name[0:-1]
                    field_name = field_name.decode("utf-8", "replace")

                    key_table.append(field_name)
                    break

        # Rewind to the beginning of the file after the header:
        # we can now parse the rest of the file.
        fp.seek(offset)

    def apps_iter():
        while True:
            appid = uint32.unpack(fp.read(4))[0]

            if appid == 0:
                break

            app = {
                'appid': appid,
                'size': uint32.unpack(fp.read(4))[0],
                'info_state': uint32.unpack(fp.read(4))[0],
                'last_updated': uint32.unpack(fp.read(4))[0],
                'access_token': uint64.unpack(fp.read(8))[0],
                'sha1': fp.read(20),
                'change_number': uint32.unpack(fp.read(4))[0],
            }

            if magic != b"'DV\x07":
                app['data_sha1'] = fp.read(20)

            # 'key_table' will be None for older 'appinfo.vdf' files which
            # use self-contained binary VDFs.
            app['data'] = binary_load(fp, key_table=key_table, mapper=dict)

            yield app


    return ({
              'magic': magic,
              'universe': universe,
            },
            apps_iter()
            )

def parse_packageinfo(fp, mapper=dict):
    """Parse packageinfo.vdf from the Steam appcache folder

    :param fp: file-like object
    :param mapper: Python object class to return
    :raises: SyntaxError
    :rtype: (:class:`Generator` returning :class:`dict` by default or mapper class if set)
    :return: (header, packages iterator)
    """
# format:
#   uint32   - MAGIC: b"'UV\x06" or b"(UV\x06"
#   uint32   - UNIVERSE: 1
#   ---- repeated package sections ----
#   uint32   - PackageID
#   20bytes  - SHA1
#   uint32   - changeNumber
#   uint64   - token           (only on b"(UV\x06")
#   variable - binary_vdf
#   ---- end of section ---------
#   uint32   - EOF: 0xFFFFFFFF

    magic = fp.read(4)
    if magic not in (b"'UV\x06", b"(UV\x06"):
        raise SyntaxError("Invalid magic, got %s" % repr(magic))

    universe = uint32.unpack(fp.read(4))[0]

    def pkgs_iter():
        while True:
            packageid = uint32.unpack(fp.read(4))[0]

            if packageid == 0xFFFFFFFF:
                break

            pkg = {
                'packageid': packageid,
                'sha1': fp.read(20),
                'change_number': uint32.unpack(fp.read(4))[0],
            }

            if magic == b"(UV\x06":
                pkg['token'] = uint64.unpack(fp.read(8))[0]

            pkg['data'] = binary_load(fp, mapper=mapper)

            yield pkg


    return ({
              'magic': magic,
              'universe': universe,
            },
            pkgs_iter()
            )


import logging
import os

from hhd.plugins.plugin import (
    Context,
    expanduser,
    fix_perms,
    get_context,
    is_steam_gamepad_running,
    restore_priviledge,
    run_steam_command,
    switch_priviledge,
)

logger = logging.getLogger(__name__)

DISTRO_NAMES = ("manjaro", "bazzite", "ubuntu", "arch")


def get_distro_color():
    match get_os():
        case "manjaro":
            return 115
        case "bazzite":
            return 265
        case "arch":
            return 195
        case "ubuntu":
            return 340
        case _:
            return 30


def get_os() -> str:
    if name := os.environ.get("HHD_DISTRO", None):
        logger.warning(f"Distro override using an environment variable to '{name}'.")
        return name

    try:
        with open("/etc/os-release") as f:
            os_release = f.read().strip().lower()
    except Exception as e:
        logger.error(f"Could not read os information, error:\n{e}")
        return "ukn"

    for name in DISTRO_NAMES:
        if name in os_release:
            logger.info(f"Running under Linux distro '{name}'.")
            return name

    logger.info(f"Running under an unknown Linux distro.")
    return "ukn"


def get_ac_status_fn() -> str | None:
    BASE_DIR = "/sys/class/power_supply"
    fn = None
    try:
        for name in os.listdir(BASE_DIR):
            if name.startswith("AC") or name.startswith("ADP"):
                fn = name
                break
        if fn is None:
            logger.error(
                f"Could not find AC status file. Power supply directory:\n{os.listdir(BASE_DIR)}"
            )
            return None

        return os.path.join(BASE_DIR, fn, "online")
    except Exception as e:
        logger.error(f"Could not read power supply directory, error:\n{e}")
        return None


def get_ac_status(fn: str | None) -> bool | None:
    if fn is None:
        return None
    if not os.path.exists(fn):
        return None
    try:
        with open(fn) as f:
            return f.read().strip() != "Discharging"
    except Exception as e:
        return None


__all__ = [
    "get_os",
    "is_steam_gamepad_running",
    "fix_perms",
    "expanduser",
    "restore_priviledge",
    "switch_priviledge",
    "get_context",
    "Context",
    "run_steam_command",
    "get_ac_status",
    "get_ac_status_fn",
]



from typing import Sequence, Mapping, Any

KEYWORDS = ["title", "hint", "options"]


def parse(data: Any, names: Sequence[str] = []):
    if not isinstance(data, Mapping):
        return []

    if "sections" in data and isinstance(data["sections"], Mapping):
        return [
            (0, None, v, [f"Section Name for: {k}"])
            for k, v in data["sections"].items()
        ]

    out = []
    fun_name = ".".join(names)
    comment = f"Setting: '{fun_name}'"
    messages = []
    if "title" in data:
        comment = f"Setting: {data['title']}"
        messages.append(("Field: title", data["title"]))
    if "hint" in data:
        messages.append(("Field: hint", data["hint"]))

    if "options" in data and isinstance(data["options"], Mapping):
        messages.extend([(f"Option: {k}", v) for k, v in data["options"].items()])

    for field, msg in messages:
        out.append((0, None, msg, [comment, field]))

    for k, v in data.items():
        out.extend(parse(v, [*names, k] if k not in ("modes", "children") else names))

    return out


def extract_hhd_yaml(
    fileobj,
    keywords: Sequence[str] = [],
    comment_tags: Sequence[str] = [],
    options: Mapping[str, Any] = {},
):
    """Extract messages from XXX files.

    :param fileobj: the file-like object the messages should be extracted
                    from
    :param keywords: a list of keywords (i.e. function names) that should
                     be recognized as translation functions
    :param comment_tags: a list of translator tags to search for and
                         include in the results
    :param options: a dictionary of additional options (optional)
    :return: an iterator over ``(lineno, funcname, message, comments)``
             tuples
    :rtype: ``iterator``
    """
    import yaml

    return parse(yaml.safe_load(fileobj))
